from django.conf.urls import url
from .views import run_scheduled_report


urlpatterns = [
    url(r'^report/(?P<pk>\d+)/run_scheduled_report/$', run_scheduled_report, name="run_scheduled_report"),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-11-21 03:32
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('report_builder', '0004_auto_20170915_2046'),
    ]

    operations = [
        migrations.AddField(
            model_name='filterfield',
            name='filter_delta',
            field=models.BigIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='filterfield',
            name='filter_type',
            field=models.CharField(blank=True, choices=[('exact', 'Equals'), ('iexact', 'Equals (case-insensitive)'), ('contains', 'Contains'), ('icontains', 'Contains (case-insensitive)'), ('in', 'in (comma seperated 1,2,3)'), ('gt', 'Greater than'), ('gte', 'Greater than equals'), ('lt', 'Less than'), ('lte', 'Less than equals'), ('startswith', 'Starts with'), ('istartswith', 'Starts with (case-insensitive)'), ('endswith', 'Ends with'), ('iendswith', 'Ends with  (case-insensitive)'), ('range', 'range'), ('relative_range', 'relative_range'), ('week_day', 'Week day'), ('isnull', 'Is null'), ('regex', 'Regular Expression'), ('iregex', 'Reg. Exp. (case-insensitive)'), ('max', 'Max (annotation-filter)'), ('min', 'Min (annotation-filter)')], default='icontains', max_length=20),
        ),
    ]



from decimal import Decimal
from itertools import chain
from numbers import Number
from django.contrib.contenttypes.models import ContentType
from django.db.models.fields import FieldDoesNotExist
from django.conf import settings
import copy
import datetime
import inspect


def javascript_date_format(python_date_format):
    format = python_date_format.replace(r'Y', 'yyyy')
    format = format.replace(r'm', 'mm')
    format = format.replace(r'd', 'dd')
    if not format:
        format = 'yyyy-mm-dd'
    return format


def duplicate(obj, changes=None):
    """ Duplicates any object including m2m fields
    changes: any changes that should occur, example
    changes = (('fullname','name (copy)'), ('do not copy me', ''))"""
    if not obj.pk:
        raise ValueError('Instance must be saved before it can be cloned.')
    duplicate = copy.copy(obj)
    duplicate.pk = None
    for change in changes:
        duplicate.__setattr__(change[0], change[1])
    duplicate.save()
    # trick to copy ManyToMany relations.
    for field in obj._meta.many_to_many:
        source = getattr(obj, field.attname)
        destination = getattr(duplicate, field.attname)
        for item in source.all():
            try:  # m2m, through fields will fail.
                destination.add(item)
            except:
                pass
    return duplicate


DATE = 1
NUMBER = 2


def sort_helper(x, sort_key, sort_type):
    """ Sadly python 3 makes it very hard to sort mixed types
        We can work around this by forcing the types
    """
    result = x[sort_key]
    if result is None:
        if sort_type == DATE:
            result = datetime.date(datetime.MINYEAR, 1, 1)
        elif sort_type == NUMBER:
            result = 0
        else:  # Last try - make it a string
            result = ''
    return result


def sort_data(data_list, display_field):
    """ Sort data based on display_field settings
    data_list - 2d array of data
    display_field - report_builder.DisplayField object
    returns sorted data_list
    """
    position = display_field.position
    is_reverse = display_field.sort_reverse
    # Try to inspect sample data to determine type
    sample_data = data_list[0][position]
    if sample_data is None:
        sample_data = data_list[-1][position]
    sort_type = None
    if isinstance(sample_data, (datetime.date, datetime.datetime)):
        sort_type = DATE
    elif isinstance(sample_data, (int, float, complex)):
        sort_type = NUMBER
    return sorted(
        data_list,
        key=lambda x: sort_helper(x, position, sort_type),
        reverse=is_reverse
    )


def increment_total(display_field, data_row):
    val = data_row[display_field.position]
    if isinstance(val, bool):
        # True: 1, False: 0
        display_field.total_count += Decimal(val)
    elif isinstance(val, Number):
        display_field.total_count += Decimal(str(val))
    elif val:
        display_field.total_count += Decimal(1)


def formatter(value, style):
    """ Convert value to Decimal to apply numeric formats.
    value - The value we wish to format.
    style - report_builder.Format object
    """
    try:
        value = Decimal(value)
    except Exception:
        pass

    try:
        return style.string.format(value)
    except ValueError:
        return value


# Model Utils


def isprop(v):
    return isinstance(v, property)


def get_properties_from_model(model_class):
    """ Show properties from a model """
    properties = []
    attr_names = [name for (name, value) in inspect.getmembers(model_class, isprop)]
    for attr_name in attr_names:
        if attr_name.endswith('pk'):
            attr_names.remove(attr_name)
        else:
            properties.append(dict(label=attr_name, name=attr_name.strip('_').replace('_', ' ')))
    return sorted(properties, key=lambda k: k['label'])


def get_relation_fields_from_model(model_class):
    """ get related fields (m2m, fk, and reverse fk) """
    relation_fields = []
    all_fields_names = get_all_field_names(model_class)
    for field_name in all_fields_names:
        field = copy.deepcopy(model_class._meta.get_field(field_name))
        direct = field.concrete
        m2m = field.many_to_many
        # get_all_field_names will return the same field
        # both with and without _id. ignore the duplicate.
        if field_name[-3:] == '_id' and field_name[:-3] in all_fields_names:
            continue
        if m2m or not direct or field.is_relation:
            field.field_name = field_name
            relation_fields += [field]
    return relation_fields


def get_all_field_names(model_class):
    """ Restores a function from django<1.10 """
    return list(set(chain.from_iterable(
        (field.name, field.attname) if hasattr(field, 'attname') else (field.name,)
        for field in model_class._meta.get_fields()
        # For complete backwards compatibility, you may want to exclude
        # GenericForeignKey from the results.
        if not (field.many_to_one and field.related_model is None)
    )))


def get_direct_fields_from_model(model_class):
    """ Direct, not m2m, not FK """
    direct_fields = []
    all_fields_names = get_all_field_names(model_class)
    for field_name in all_fields_names:
        field = model_class._meta.get_field(field_name)
        direct = field.concrete
        m2m = field.many_to_many
        if direct and not m2m and not field.is_relation:
            direct_fields += [field]
    return direct_fields


def get_custom_fields_from_model(model_class):
    """ django-custom-fields support """
    if 'custom_field' in settings.INSTALLED_APPS:
        from custom_field.models import CustomField
        try:
            content_type = ContentType.objects.get(
                model=model_class._meta.model_name,
                app_label=model_class._meta.app_label)
        except ContentType.DoesNotExist:
            content_type = None
        custom_fields = CustomField.objects.filter(content_type=content_type)
        return custom_fields


def get_model_from_path_string(root_model, path):
    """ Return a model class for a related model
    root_model is the class of the initial model
    path is like foo__bar where bar is related to foo
    """
    for path_section in path.split('__'):
        if path_section:
            try:
                field = root_model._meta.get_field(path_section)
                direct = field.concrete
            except FieldDoesNotExist:
                return root_model
            if direct:
                if hasattr(field, 'related'):
                    try:
                        root_model = field.related.parent_model()
                    except AttributeError:
                        root_model = field.related.model

                elif hasattr(field, 'related_model') and field.related_model:
                    root_model = field.related_model

            else:
                if hasattr(field, 'related_model'):
                    root_model = field.related_model
                else:
                    root_model = field.model
    return root_model



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('report_builder', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='displayfield',
            name='aggregate',
            field=models.CharField(choices=[('Sum', 'Sum'), ('Count', 'Count'), ('Avg', 'Avg'), ('Max', 'Max'), ('Min', 'Min')], blank=True, max_length=5),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='displayfield',
            name='sort_reverse',
            field=models.BooleanField(default=False, verbose_name='Reverse'),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='filterfield',
            name='filter_type',
            field=models.CharField(blank=True, default='icontains', choices=[('exact', 'Equals'), ('iexact', 'Equals (case-insensitive)'), ('contains', 'Contains'), ('icontains', 'Contains (case-insensitive)'), ('in', 'in (comma seperated 1,2,3)'), ('gt', 'Greater than'), ('gte', 'Greater than equals'), ('lt', 'Less than'), ('lte', 'Less than equals'), ('startswith', 'Starts with'), ('istartswith', 'Starts with (case-insensitive)'), ('endswith', 'Ends with'), ('iendswith', 'Ends with  (case-insensitive)'), ('range', 'range'), ('week_day', 'Week day'), ('isnull', 'Is null'), ('regex', 'Regular Expression'), ('iregex', 'Reg. Exp. (case-insensitive)')], max_length=20),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='format',
            name='name',
            field=models.CharField(blank=True, default='', max_length=50),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='format',
            name='string',
            field=models.CharField(blank=True, default='', help_text='Python string format. Ex ${} would place a $ in front of the result.', max_length=300),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='report',
            name='report_file',
            field=models.FileField(upload_to='report_files', blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='report',
            name='slug',
            field=models.SlugField(verbose_name='Short Name'),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='report',
            name='starred',
            field=models.ManyToManyField(help_text='These users have starred this report for easy reference.', blank=True, related_name='report_starred_set', to=settings.AUTH_USER_MODEL),
            preserve_default=True,
        ),
    ]



from django.conf import settings
from django.template.loader import get_template
from django.core.mail import send_mail, EmailMultiAlternatives


def email_report(report_url, user=None, email=None):
    if ((getattr(settings, 'EMAIL_BACKEND', False) or
             getattr(settings, 'EMAIL_HOST', False)) and
            getattr(settings, 'DEFAULT_FROM_EMAIL', False)):

        name = None
        if user:
            email = user.email
            name = user.username

        if get_template('email/email_report.html'):
            email_template = get_template('email/email_report.html')
            msg = EmailMultiAlternatives(
                getattr(settings, 'REPORT_BUILDER_EMAIL_SUBJECT', False) or
                'Report is ready',
                report_url,
                getattr(settings, 'DEFAULT_FROM_EMAIL'),
                [email],
            )
            htmlParameters = {
                'name': name,
                'report': report_url,
            }
            msg.attach_alternative(
                email_template.render(htmlParameters),
                "text/html"
            )
            msg.send()
        else:
            send_mail(
                getattr(settings, 'REPORT_BUILDER_EMAIL_SUBJECT', False) or
                'Report is ready',
                str(report_url),
                getattr(settings, 'DEFAULT_FROM_EMAIL'),
                [email],
                fail_silently=True,
            )



from six import BytesIO, StringIO, text_type, string_types

from django.http import HttpResponse
from django.contrib.contenttypes.models import ContentType
try:
    from django.db.models.fields.related_descriptors import ManyToManyDescriptor
except ImportError:
    # Django 1.8 compat hack.
    from django.db.models.fields.related import (
        ReverseManyRelatedObjectsDescriptor as ManyToManyDescriptor
    )
from django.db.models import Avg, Count, Sum, Max, Min
from openpyxl.workbook import Workbook
from openpyxl.writer.excel import save_virtual_workbook
from openpyxl.utils import get_column_letter
from openpyxl.styles import Font
import csv
import re
from collections import namedtuple
from decimal import Decimal
from numbers import Number
from functools import reduce
import datetime

from .utils import (
    get_relation_fields_from_model,
    get_properties_from_model,
    get_direct_fields_from_model,
    get_model_from_path_string,
    get_custom_fields_from_model,
)


DisplayField = namedtuple(
    "DisplayField",
    "path path_verbose field field_verbose aggregate total group choices field_type",
)


def generate_filename(title, ends_with):
    title = title.split('.')[0]
    title.replace(' ', '_')
    title += ('_' + datetime.datetime.now().strftime("%m%d_%H%M"))
    if not title.endswith(ends_with):
        title += ends_with
    return title


class DataExportMixin(object):
    def build_sheet(self, data, ws, sheet_name='report', header=None, widths=None):
        first_row = 1
        column_base = 1

        ws.title = re.sub(r'\W+', '', sheet_name)[:30]
        if header:
            for i, header_cell in enumerate(header):
                cell = ws.cell(row=first_row, column=i + column_base)
                cell.value = header_cell
                cell.font = Font(bold=True)
                if widths:
                    ws.column_dimensions[get_column_letter(i + 1)].width = widths[i]

        for row in data:
            for i in range(len(row)):
                item = row[i]
                # If item is a regular string
                if isinstance(item, str):
                    # Change it to a unicode string
                    try:
                        row[i] = text_type(item)
                    except UnicodeDecodeError:
                        row[i] = text_type(item.decode('utf-8', 'ignore'))
                elif type(item) is dict:
                    row[i] = text_type(item)
            try:
                ws.append(row)
            except ValueError as e:
                ws.append([e.message])
            except:
                ws.append(['Unknown Error'])

    def build_xlsx_response(self, wb, title="report"):
        """ Take a workbook and return a xlsx file response """
        title = generate_filename(title, '.xlsx')
        myfile = BytesIO()
        myfile.write(save_virtual_workbook(wb))
        response = HttpResponse(
            myfile.getvalue(),
            content_type='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet')
        response['Content-Disposition'] = 'attachment; filename=%s' % title
        response['Content-Length'] = myfile.tell()
        return response

    def build_csv_response(self, wb, title="report"):
        """ Take a workbook and return a csv file response """
        title = generate_filename(title, '.csv')
        myfile = StringIO()
        sh = wb.active
        c = csv.writer(myfile)
        for r in sh.rows:
            c.writerow([cell.value for cell in r])
        response = HttpResponse(
            myfile.getvalue(),
            content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename=%s' % title
        response['Content-Length'] = myfile.tell()
        return response

    def list_to_workbook(self, data, title='report', header=None, widths=None):
        """ Create just a openpxl workbook from a list of data """
        wb = Workbook()
        title = re.sub(r'\W+', '', title)[:30]

        if isinstance(data, dict):
            i = 0
            for sheet_name, sheet_data in data.items():
                if i > 0:
                    wb.create_sheet()
                ws = wb.worksheets[i]
                self.build_sheet(
                    sheet_data, ws, sheet_name=sheet_name, header=header)
                i += 1
        else:
            ws = wb.worksheets[0]
            self.build_sheet(data, ws, header=header, widths=widths)
        return wb

    def list_to_xlsx_file(self, data, title='report', header=None, widths=None):
        """ Make 2D list into a xlsx response for download
        data can be a 2d array or a dict of 2d arrays
        like {'sheet_1': [['A1', 'B1']]}
        returns a StringIO file
        """
        wb = self.list_to_workbook(data, title, header, widths)
        if not title.endswith('.xlsx'):
            title += '.xlsx'
        myfile = BytesIO()
        myfile.write(save_virtual_workbook(wb))
        return myfile

    def list_to_csv_file(self, data, title='report', header=None, widths=None):
        """ Make a list into a csv response for download.
        """
        wb = self.list_to_workbook(data, title, header, widths)
        if not title.endswith('.csv'):
            title += '.csv'
        myfile = StringIO()
        sh = wb.active
        c = csv.writer(myfile)
        for r in sh.rows:
            c.writerow([cell.value for cell in r])
        return myfile

    def list_to_xlsx_response(self, data, title='report', header=None,
                              widths=None):
        """ Make 2D list into a xlsx response for download
        data can be a 2d array or a dict of 2d arrays
        like {'sheet_1': [['A1', 'B1']]}
        """
        wb = self.list_to_workbook(data, title, header, widths)
        return self.build_xlsx_response(wb, title=title)

    def list_to_csv_response(self, data, title='report', header=None,
                             widths=None):
        """ Make 2D list into a csv response for download data.
        """
        wb = self.list_to_workbook(data, title, header, widths)
        return self.build_csv_response(wb, title=title)

    def add_aggregates(self, queryset, display_fields):
        agg_funcs = {
            'Avg': Avg, 'Min': Min, 'Max': Max, 'Count': Count, 'Sum': Sum
        }

        for display_field in display_fields:
            if display_field.aggregate:
                func = agg_funcs[display_field.aggregate]
                full_name = display_field.path + display_field.field
                queryset = queryset.annotate(func(full_name))

        return queryset

    def report_to_list(self, queryset, display_fields, user=None, property_filters=[], preview=False):
        """ Create list from a report with all data filtering.
        queryset: initial queryset to generate results
        display_fields: list of field references or DisplayField models
        user: requesting user. If left as None - there will be no permission check
        property_filters: ???
        preview: return only first 50 rows
        Returns list, message in case of issues.
        """
        model_class = queryset.model

        def can_change_or_view(model):
            """ Return True iff `user` has either change or view permission
            for `model`. """
            if user is None:
                return True
            model_name = model._meta.model_name
            app_label = model._meta.app_label
            can_change = user.has_perm(app_label + '.change_' + model_name)
            can_view = user.has_perm(app_label + '.view_' + model_name)

            return can_change or can_view

        if not can_change_or_view(model_class):
            return [], 'Permission Denied'

        if isinstance(display_fields, list):
            # Convert list of strings to DisplayField objects.

            new_display_fields = []

            for display_field in display_fields:
                field_list = display_field.split('__')
                field = field_list[-1]
                path = '__'.join(field_list[:-1])

                if path:
                    path += '__'  # Legacy format to append a __ here.

                new_model = get_model_from_path_string(model_class, path)
                model_field = new_model._meta.get_field_by_name(field)[0]
                choices = model_field.choices
                new_display_fields.append(DisplayField(
                    path, '', field, '', '', None, None, choices, ''
                ))

            display_fields = new_display_fields

        # Build group-by field list.

        group = [df.path + df.field for df in display_fields if df.group]

        # To support group-by with multiple fields, we turn all the other
        # fields into aggregations. The default aggregation is `Max`.

        if group:
            for field in display_fields:
                if (not field.group) and (not field.aggregate):
                    field.aggregate = 'Max'

        message = ""
        objects = self.add_aggregates(queryset, display_fields)

        # Display Values

        display_field_paths = []
        property_list = {}
        custom_list = {}
        display_totals = {}

        for i, display_field in enumerate(display_fields):
            model = get_model_from_path_string(model_class, display_field.path)

            if display_field.field_type == "Invalid":
                continue

            if not model or can_change_or_view(model):
                display_field_key = display_field.path + display_field.field

                if display_field.field_type == "Property":
                    property_list[i] = display_field_key
                elif display_field.field_type == "Custom Field":
                    custom_list[i] = display_field_key
                elif display_field.aggregate == "Avg":
                    display_field_key += '__avg'
                elif display_field.aggregate == "Max":
                    display_field_key += '__max'
                elif display_field.aggregate == "Min":
                    display_field_key += '__min'
                elif display_field.aggregate == "Count":
                    display_field_key += '__count'
                elif display_field.aggregate == "Sum":
                    display_field_key += '__sum'

                if display_field.field_type not in ('Property', 'Custom Field'):
                    display_field_paths.append(display_field_key)

                if display_field.total:
                    display_totals[display_field_key] = Decimal(0)

            else:
                message += 'Error: Permission denied on access to {0}.'.format(
                    display_field.name
                )

        def increment_total(display_field_key, val):
            """ Increment display total by `val` if given `display_field_key` in
            `display_totals`.
            """
            if display_field_key in display_totals:
                if isinstance(val, bool):
                    # True: 1, False: 0
                    display_totals[display_field_key] += Decimal(val)
                elif isinstance(val, Number):
                    display_totals[display_field_key] += Decimal(str(val))
                elif val:
                    display_totals[display_field_key] += Decimal(1)

        # Select pk for primary and m2m relations in order to retrieve objects
        # for adding properties to report rows. Group-by queries do not support
        # Property nor Custom Field filters.

        if not group:
            display_field_paths.insert(0, 'pk')

            m2m_relations = []
            for position, property_path in property_list.items():
                property_root = property_path.split('__')[0]
                root_class = model_class

                try:
                    property_root_class = getattr(root_class, property_root)
                except AttributeError:  # django-hstore schema compatibility
                    continue

                if type(property_root_class) == ManyToManyDescriptor:
                    display_field_paths.insert(1, '%s__pk' % property_root)
                    m2m_relations.append(property_root)

        if group:
            values = objects.values(*group)
            values = self.add_aggregates(values, display_fields)
            filtered_report_rows = [
                [row[field] for field in display_field_paths]
                for row in values
            ]
            for row in filtered_report_rows:
                for pos, field in enumerate(display_field_paths):
                    increment_total(field, row[pos])
        else:
            filtered_report_rows = []
            values_and_properties_list = []

            values_list = objects.values_list(*display_field_paths)

            for row in values_list:
                row = list(row)
                values_and_properties_list.append(row[1:])
                obj = None  # we will get this only if needed for more complex processing
                # related_objects
                remove_row = False
                # filter properties (remove rows with excluded properties)
                for property_filter in property_filters:
                    if not obj:
                        obj = model_class.objects.get(pk=row.pop(0))
                    root_relation = property_filter.path.split('__')[0]
                    if root_relation in m2m_relations:
                        pk = row[0]
                        if pk is not None:
                            # a related object exists
                            m2m_obj = getattr(obj, root_relation).get(pk=pk)
                            val = reduce(getattr, [property_filter.field], m2m_obj)
                        else:
                            val = None
                    else:
                        if property_filter.field_type == 'Custom Field':
                            for relation in property_filter.path.split('__'):
                                if hasattr(obj, root_relation):
                                    obj = getattr(obj, root_relation)
                            val = obj.get_custom_value(property_filter.field)
                        else:
                            val = reduce(getattr, (property_filter.path + property_filter.field).split('__'), obj)
                    if property_filter.filter_property(val):
                        remove_row = True
                        values_and_properties_list.pop()
                        break
                if not remove_row:
                    for i, field in enumerate(display_field_paths[1:]):
                        increment_total(field, row[i + 1])

                    for position, display_property in property_list.items():
                        if not obj:
                            obj = model_class.objects.get(pk=row.pop(0))
                        relations = display_property.split('__')
                        root_relation = relations[0]
                        if root_relation in m2m_relations:
                            pk = row.pop(0)
                            if pk is not None:
                                # a related object exists
                                m2m_obj = getattr(obj, root_relation).get(pk=pk)
                                val = reduce(getattr, relations[1:], m2m_obj)
                            else:
                                val = None
                        else:
                            # Could error if a related field doesn't exist
                            try:
                                val = reduce(getattr, relations, obj)
                            except AttributeError:
                                val = None
                        values_and_properties_list[-1].insert(position, val)
                        increment_total(display_property, val)

                    for position, display_custom in custom_list.items():
                        if not obj:
                            obj = model_class.objects.get(pk=row.pop(0))
                        val = obj.get_custom_value(display_custom)
                        values_and_properties_list[-1].insert(position, val)
                        increment_total(display_custom, val)

                    filtered_report_rows.append(values_and_properties_list[-1])

                if preview and len(filtered_report_rows) == 50:
                    break

        # Sort results if requested.

        if hasattr(display_fields, 'filter'):
            defaults = {
                None: text_type,
                datetime.date: lambda: datetime.date(datetime.MINYEAR, 1, 1),
                datetime.datetime: lambda: datetime.datetime(datetime.MINYEAR, 1, 1),
            }

            # Order sort fields in reverse order so that ascending, descending
            # sort orders work together (based on Python's stable sort). See
            # http://stackoverflow.com/questions/6666748/ for details.

            sort_fields = display_fields.filter(sort__gt=0).order_by('-sort')
            sort_values = sort_fields.values_list('position', 'sort_reverse')

            for pos, reverse in sort_values:
                column = (row[pos] for row in filtered_report_rows)
                type_col = (type(val) for val in column if val is not None)
                field_type = next(type_col, None)
                default = defaults.get(field_type, field_type)()

                filtered_report_rows = sorted(
                    filtered_report_rows,
                    key=lambda row: self.sort_helper(row[pos], default),
                    reverse=reverse,
                )

        values_and_properties_list = filtered_report_rows

        # Build mapping from display field position to choices list.

        choice_lists = {}
        for df in display_fields:
            if df.choices and hasattr(df, 'choices_dict'):
                df_choices = df.choices_dict
                # Insert blank and None as valid choices.
                df_choices[''] = ''
                df_choices[None] = ''
                choice_lists[df.position] = df_choices

        # Build mapping from display field position to format.

        display_formats = {}

        for df in display_fields:
            if hasattr(df, 'display_format') and df.display_format:
                display_formats[df.position] = df.display_format

        def formatter(value, style):
            # Convert value to Decimal to apply numeric formats.
            try:
                value = Decimal(value)
            except Exception:
                pass

            try:
                return style.string.format(value)
            except ValueError:
                return value

        # Iterate rows and convert values by choice lists and field formats.

        final_list = []

        for row in values_and_properties_list:
            row = list(row)

            for position, choice_list in choice_lists.items():
                try:
                    row[position] = text_type(choice_list[row[position]])
                except Exception:
                    row[position] = text_type(row[position])

            for pos, style in display_formats.items():
                row[pos] = formatter(row[pos], style)

            final_list.append(row)

        values_and_properties_list = final_list

        if display_totals:
            display_totals_row = []

            fields_and_properties = list(display_field_paths[0 if group else 1:])

            for position, value in property_list.items():
                fields_and_properties.insert(position, value)

            for field in fields_and_properties:
                display_totals_row.append(display_totals.get(field, ''))

            # Add formatting to display totals.

            for pos, style in display_formats.items():
                display_totals_row[pos] = formatter(display_totals_row[pos], style)

            values_and_properties_list.append(
                ['TOTALS'] + (len(fields_and_properties) - 1) * ['']
            )
            values_and_properties_list.append(display_totals_row)

        return values_and_properties_list, message

    def sort_helper(self, value, default):
        if value is None:
            value = default
        if isinstance(value, string_types):
            value = value.lower()
        return value


class GetFieldsMixin(object):
    def get_fields(self, model_class, field_name='', path='', path_verbose=''):
        """ Get fields and meta data from a model
        :param model_class: A django model class
        :param field_name: The field name to get sub fields from
        :param path: path of our field in format
            field_name__second_field_name__ect__
        :param path_verbose: Human readable version of above
        :returns: Returns fields and meta data about such fields
            fields: Django model fields
            custom_fields: fields from django-custom-field if installed
            properties: Any properties the model has
            path: Our new path
            path_verbose: Our new human readable path
        :rtype: dict
        """
        fields = get_direct_fields_from_model(model_class)
        properties = get_properties_from_model(model_class)
        custom_fields = get_custom_fields_from_model(model_class)
        app_label = model_class._meta.app_label

        if field_name != '':
            field = model_class._meta.get_field(field_name)
            direct = field.concrete
            if path_verbose:
                path_verbose += "::"
            # TODO: need actual model name to generate choice list (not pluralized field name)
            # - maybe store this as a separate value?
            if field.many_to_many and hasattr(field, 'm2m_reverse_field_name'):
                path_verbose += field.m2m_reverse_field_name()
            else:
                path_verbose += field.name

            path += field_name
            path += '__'
            if direct:
                new_model = field.related_model
                path_verbose = new_model.__name__.lower()
            else:  # Indirect related field
                new_model = field.related_model
                path_verbose = new_model.__name__.lower()

            fields = get_direct_fields_from_model(new_model)

            custom_fields = get_custom_fields_from_model(new_model)
            properties = get_properties_from_model(new_model)
            app_label = new_model._meta.app_label

        return {
            'fields': fields,
            'custom_fields': custom_fields,
            'properties': properties,
            'path': path,
            'path_verbose': path_verbose,
            'app_label': app_label,
        }

    def get_related_fields(self, model_class, field_name, path="", path_verbose=""):
        """ Get fields for a given model """
        if field_name:
            field = model_class._meta.get_field(field_name)
            direct = field.concrete
            if direct:
                try:
                    related_field = field.remote_field
                except AttributeError:
                    # Needed for Django < 1.9
                    related_field = field.related
                try:
                    new_model = related_field.parent_model()
                except AttributeError:
                    new_model = related_field.model
            else:
                # Indirect related field
                new_model = field.related_model

            if path_verbose:
                path_verbose += "::"
            path_verbose += field.name

            path += field_name
            path += '__'
        else:
            new_model = model_class

        new_fields = get_relation_fields_from_model(new_model)
        model_ct = ContentType.objects.get_for_model(new_model)

        return (new_fields, model_ct, path)



from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from .tasks import report_builder_run_scheduled_report
from .models import ScheduledReport

@staff_member_required
def run_scheduled_report(request, pk):
    """ Manually run a scheduled report - useful for testing or one off situations """
    scheduled_report = get_object_or_404(ScheduledReport, pk=pk)
    report_builder_run_scheduled_report.delay(scheduled_report.id)
    messages.success(request, "Ran scheduled report")
    return HttpResponseRedirect(reverse('admin:report_builder_scheduled_scheduledreport_changelist'))



from django import forms
from django.contrib.contenttypes.models import ContentType
from django.conf import settings
from django.core.urlresolvers import reverse
from django.core.files.base import ContentFile
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.utils.safestring import mark_safe
from django.utils.functional import cached_property
from django.db import models
from django.db.models import Avg, Min, Max, Count, Sum, F
from django.db.models.fields import FieldDoesNotExist
from report_builder.unique_slugify import unique_slugify
from .utils import (
    get_model_from_path_string, sort_data, increment_total, formatter)
from .mixins import generate_filename, DataExportMixin
from .email import email_report
from dateutil import parser
from decimal import Decimal
from functools import reduce
import time
import datetime
import re
import freezegun

AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


def get_allowed_models():
    models = ContentType.objects.all()
    if getattr(settings, 'REPORT_BUILDER_INCLUDE', False):
        all_model_names = []
        additional_models = []
        for element in settings.REPORT_BUILDER_INCLUDE:
            split_element = element.split('.')
            if len(split_element) == 2:
                additional_models.append(models.filter(app_label=split_element[0], model=split_element[1]))
            else:
                all_model_names.append(element)
        models = models.filter(model__in=all_model_names)
        for additional_model in additional_models:
            models = models | additional_model
    if getattr(settings, 'REPORT_BUILDER_EXCLUDE', False):
        all_model_names = []
        for element in settings.REPORT_BUILDER_EXCLUDE:
            split_element = element.split('.')
            if len(split_element) == 2:
                models = models.exclude(app_label=split_element[0], model=split_element[1])
            else:
                all_model_names.append(element)
        models = models.exclude(model__in=all_model_names)
    return models


def get_limit_choices_to_callable():
    return {'pk__in': get_allowed_models()}


class Report(models.Model):
    """ A saved report with queryset and descriptive fields
    """
    def _get_model_manager(self):
        """ Get  manager from settings else use objects
        """
        model_manager = 'objects'
        if getattr(settings, 'REPORT_BUILDER_MODEL_MANAGER', False):
            model_manager = settings.REPORT_BUILDER_MODEL_MANAGER
        return model_manager

    @staticmethod
    def allowed_models():
        return get_allowed_models()

    name = models.CharField(max_length=255)
    slug = models.SlugField(verbose_name="Short Name")
    description = models.TextField(blank=True)
    root_model = models.ForeignKey(
        ContentType, limit_choices_to=get_limit_choices_to_callable)
    created = models.DateField(auto_now_add=True)
    modified = models.DateField(auto_now=True)
    user_created = models.ForeignKey(
        AUTH_USER_MODEL, editable=False, blank=True, null=True)
    user_modified = models.ForeignKey(
        AUTH_USER_MODEL, editable=False, blank=True, null=True,
        related_name="report_modified_set")
    distinct = models.BooleanField(default=False)
    report_file = models.FileField(upload_to="report_files", blank=True)
    report_file_creation = models.DateTimeField(blank=True, null=True)
    starred = models.ManyToManyField(
        AUTH_USER_MODEL, blank=True,
        help_text="These users have starred this report for easy reference.",
        related_name="report_starred_set")

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if not self.id:
            unique_slugify(self, self.name)
        super(Report, self).save(*args, **kwargs)

    def add_aggregates(self, queryset, display_fields=None):
        agg_funcs = {
            'Avg': Avg, 'Min': Min, 'Max': Max, 'Count': Count, 'Sum': Sum
        }
        if display_fields is None:
            display_fields = self.displayfield_set.filter(
                aggregate__isnull=False)
        for display_field in display_fields:
            if display_field.aggregate:
                func = agg_funcs[display_field.aggregate]
                full_name = display_field.path + display_field.field
                queryset = queryset.annotate(func(full_name))

        return queryset

    @property
    def root_model_class(self):
        return self.root_model.model_class()

    def get_field_type(self, field_name, path=""):
        """ Get field type for given field name.
        field_name is the full path of the field
        path is optional
        """
        model = get_model_from_path_string(
            self.root_model_class, path + field_name)

        # Is it a ORM field?
        try:
            return model._meta.get_field(field_name).get_internal_type()
        except FieldDoesNotExist:
            pass
        # Is it a property?
        field_attr = getattr(model, field_name, None)
        if isinstance(field_attr, (property, cached_property)):
            return "Property"
        # Is it a custom field?
        try:
            model().get_custom_field(field_name)
            return "Custom Field"
        except (ObjectDoesNotExist, AttributeError):
            pass
        return "Invalid"

    def get_good_display_fields(self):
        """ Returns only valid display fields """
        display_fields = self.displayfield_set.all()
        bad_display_fields = []
        for display_field in display_fields:
            if display_field.field_type == "Invalid":
                bad_display_fields.append(display_field)
        return display_fields.exclude(id__in=[o.id for o in bad_display_fields])

    def report_to_list(self, queryset=None, user=None, preview=False):
        """Convert report into list."""
        property_filters = []
        if queryset is None:
            queryset = self.get_query()
            for field in self.filterfield_set.all():
                if field.field_type in ["Property"]:
                    property_filters += [field]
        display_fields = self.get_good_display_fields()

        # Need the pk for inserting properties later
        display_field_paths = ['pk']
        display_field_properties = []
        display_totals = []
        insert_property_indexes = []
        choice_lists = {}
        display_formats = {}
        i = 0
        for display_field in display_fields:
            if display_field.total:
                display_field.total_count = Decimal(0.0)
                display_totals.append(display_field)
            display_field_type = display_field.field_type
            if display_field_type == "Property":
                display_field_properties.append(display_field.field_key)
                insert_property_indexes.append(i)
            else:
                if display_field.aggregate:
                    display_field_paths += [
                        display_field.field_key +
                        '__' + display_field.aggregate.lower()]
                else:
                    display_field_paths += [display_field.field_key]
            i += 1

            # Build display choices list
            if display_field.choices and hasattr(display_field, 'choices_dict'):
                choice_list = display_field.choices_dict
                # Insert blank and None as valid choices.
                choice_list[''] = ''
                choice_list[None] = ''
                choice_lists[display_field.position] = choice_list

            # Build display format list
            if (
                hasattr(display_field, 'display_format') and
                display_field.display_format
            ):
                display_formats[display_field.position] = \
                    display_field.display_format

        property_filters = []
        for filter_field in self.filterfield_set.all():
            filter_field_type = filter_field.field_type
            if filter_field_type == "Property":
                property_filters += [field]

        group = [df.path + df.field for df in display_fields if df.group]

        # To support group-by with multiple fields, we turn all the other
        # fields into aggregations. The default aggregation is `Max`.
        if group:
            for field in display_fields:
                if (not field.group) and (not field.aggregate):
                    field.aggregate = 'Max'
            values = queryset.values(*group)
            values = self.add_aggregates(values, display_fields)
            data_list = []
            for row in values:
                row_data = []
                for field in display_field_paths:
                    if field == 'pk':
                        continue
                    try:
                        row_data.append(row[field])
                    except KeyError:
                        row_data.append(row[field + '__max'])
                for total in display_totals:
                    increment_total(total, row_data)
                data_list.append(row_data)
        else:
            values_list = list(queryset.values_list(*display_field_paths))

            data_list = []
            values_index = 0
            for obj in queryset:
                display_property_values = []
                for display_property in display_field_properties:
                    relations = display_property.split('__')
                    val = reduce(getattr, relations, obj)
                    display_property_values.append(val)

                value_row = values_list[values_index]
                while value_row[0] == obj.pk:
                    add_row = True
                    data_row = list(value_row[1:])  # Remove added pk
                    # Insert in the location dictated by the order of display fields
                    for i, prop_value in enumerate(display_property_values):
                        data_row.insert(insert_property_indexes[i], prop_value)
                    for property_filter in property_filters:
                        relations = property_filter.field_key.split('__')
                        val = reduce(getattr, relations, obj)
                        if property_filter.filter_property(val):
                            add_row = False

                    if add_row is True:
                        for total in display_totals:
                            increment_total(total, data_row)
                        # Replace choice data with display choice string
                        for position, choice_list in choice_lists.items():
                            try:
                                data_row[position] = str(choice_list[data_row[position]])
                            except Exception:
                                data_row[position] = str(data_row[position])
                        for position, style in display_formats.items():
                            data_row[position] = formatter(data_row[position], style)
                        data_list.append(data_row)
                    values_index += 1
                    try:
                        value_row = values_list[values_index]
                    except IndexError:
                        break

        for display_field in display_fields.filter(
            sort__gt=0
        ).order_by('-sort'):
            data_list = sort_data(data_list, display_field)

        if display_totals:
            display_totals_row = []
            i = 0
            for display_field in display_totals:
                while i < display_field.position:
                    i += 1
                    display_totals_row.append('')
                i += 1
                display_totals_row.append(display_field.total_count)
            # Add formats to display totals
            for pos, style in display_formats.items():
                display_totals_row[pos] = formatter(display_totals_row[pos], style)

            data_list += [
                ['TOTALS'] + (len(display_fields) - 1) * ['']
            ] + [display_totals_row]

        return data_list

    def get_query(self):
        report = self
        model_class = self.root_model_class

        # Check for report_builder_model_manger property on the model
        if getattr(model_class, 'report_builder_model_manager', False):
            objects = getattr(model_class, 'report_builder_model_manager').all()
        else:
            # Get global model manager
            manager = report._get_model_manager()
            objects = getattr(model_class, manager).all()

        # Filters
        # NOTE: group all the filters together into one in order to avoid
        # unnecessary joins
        filters = {}
        excludes = {}
        for filter_field in report.filterfield_set.order_by('position'):
            # exclude properties from standard ORM filtering
            if filter_field.field_type == "Property":
                continue
            if filter_field.field_type == "Custom Field":
                continue
            if filter_field.filter_type in ('max', 'min'):
                # Annotation-filters are handled below.
                continue

            filter_string = str(filter_field.path + filter_field.field)

            if filter_field.filter_type:
                ft = filter_field.filter_type
                fs = ft if ft != 'relative_range' else 'range'
                filter_string += '__' + fs

            # Check for special types such as isnull
            if (filter_field.filter_type == "isnull" and
               filter_field.filter_value in ["0", "False"]):
                filter_ = {filter_string: False}
            elif filter_field.filter_type == "in":
                filter_ = {filter_string: filter_field.filter_value.split(',')}
            else:
                # Check for range and relative_range types
                filter_value = filter_field.filter_value
                if filter_field.filter_type == 'range':
                    filter_value = sorted([filter_value, filter_field.filter_value2])
                elif filter_field.filter_type == 'relative_range':
                    filter_value = filter_field.get_relative_range()
                filter_ = {filter_string: filter_value}

            if not filter_field.exclude:
                filters.update(filter_)
            else:
                excludes.update(filter_)

        if filters:
            objects = objects.filter(**filters)
        if excludes:
            objects = objects.exclude(**excludes)

        # Apply annotation-filters after regular filters.
        for filter_field in report.filterfield_set.order_by('position'):
            if filter_field.filter_type in ('max', 'min'):
                func = {'max': Max, 'min': Min}[filter_field.filter_type]
                column_name = '{0}{1}__{2}'.format(
                    filter_field.path,
                    filter_field.field,
                    filter_field.field_type
                )
                filter_string = filter_field.path + filter_field.field
                annotate_args = {column_name: func(filter_string)}
                filter_args = {column_name: F(filter_field.field)}
                objects = objects.annotate(**annotate_args).filter(**filter_args)

        # Aggregates
        objects = self.add_aggregates(objects)

        # Distinct
        if report.distinct:
            objects = objects.distinct()

        return objects

    @models.permalink
    def get_absolute_url(self):
        return ("report_update_view", [str(self.id)])

    def edit(self):
        return mark_safe(
            '<a href="{0}"><img style="width: 26px; margin: -6px" src="{1}report_builder/img/edit.svg"/></a>'.format(
                self.get_absolute_url(),
                getattr(settings, 'STATIC_URL', '/static/')
            )
        )
    edit.allow_tags = True

    def download_xlsx(self):
        if getattr(settings, 'REPORT_BUILDER_ASYNC_REPORT', False):
            return mark_safe(
                '<a href="javascript:void(0)" onclick="get_async_report({0})"><img style="width: 26px; margin: -6px" src="{1}report_builder/img/download.svg"/></a>'.format(
                    self.id,
                    getattr(settings, 'STATIC_URL', '/static/'),
                )
            )
        else:
            return mark_safe(
                '<a href="{0}"><img style="width: 26px; margin: -6px" src="{1}report_builder/img/download.svg"/></a>'.format(
                    reverse('report_download_file', args=[self.id]),
                    getattr(settings, 'STATIC_URL', '/static/'),
                )
            )
    download_xlsx.short_description = "Download"
    download_xlsx.allow_tags = True

    def copy_report(self):
        return '<a href="{0}"><img style="width: 26px; margin: -6px" src="{1}report_builder/img/copy.svg"/></a>'.format(
            reverse('report_builder_create_copy', args=[self.id]),
            getattr(settings, 'STATIC_URL', '/static/'),
        )
    copy_report.short_description = "Copy"
    copy_report.allow_tags = True

    def check_report_display_field_positions(self):
        """ After report is saved, make sure positions are sane
        """
        for i, display_field in enumerate(self.displayfield_set.all()):
            if display_field.position != i + 1:
                display_field.position = i + 1
                display_field.save()

    def email_report(self, user=None, email=None):
        report_url = self.report_file.url
        return email_report(report_url, user=user, email=email)

    def async_report_save(self, objects_list,
                          title, header, widths, user=None, file_type="xlsx", email_to:str = None):
        data_export = DataExportMixin()
        if file_type == 'csv':
            csv_file = data_export.list_to_csv_file(objects_list, title,
                                                    header, widths)
            title = generate_filename(title, '.csv')
            self.report_file.save(title, ContentFile(csv_file.getvalue()))
        else:
            xlsx_file = data_export.list_to_xlsx_file(objects_list, title,
                                                      header, widths)
            title = generate_filename(title, '.xlsx')
            self.report_file.save(title, ContentFile(xlsx_file.getvalue()))
        self.report_file_creation = datetime.datetime.today()
        self.save()
        if email_to:
            for email in email_to:
                self.email_report(email=email)
        elif getattr(settings, 'REPORT_BUILDER_EMAIL_NOTIFICATION', False):
            if user.email:
                self.email_report(user=user)

    def run_report(self, file_type, user=None, queryset=None, async=False, scheduled=False, email_to:str = None):
        """Generate this report file"""
        if not queryset:
            queryset = self.get_query()

        display_fields = self.get_good_display_fields()

        data_export = DataExportMixin()
        objects_list, message = data_export.report_to_list(
            queryset, display_fields, user, preview=False)
        title = re.sub(r'\W+', '', self.name)[:30]
        header = []
        widths = []
        for field in display_fields:
            header.append(field.name)
            widths.append(field.width)
        if scheduled:
            self.async_report_save(objects_list, title, header, widths, file_type, email_to=email_to)
        elif async:
            if user is None:
                raise Exception('Cannot run async report without a user')
            self.async_report_save(objects_list, title, header, widths, user, file_type)
        else:
            if file_type == 'csv':
                return data_export.list_to_csv_response(
                    objects_list, title, header, widths)
            else:
                return data_export.list_to_xlsx_response(
                    objects_list, title, header, widths)


class Format(models.Model):
    """ A specifies a Python string format for e.g. `DisplayField`s.
    """
    name = models.CharField(max_length=50, blank=True, default='')
    string = models.CharField(
        max_length=300, blank=True, default='',
        help_text='Python string format. Ex ${} would place a $ in front of the result.'
    )

    def __str__(self):
        return self.name


class AbstractField(models.Model):
    report = models.ForeignKey(Report)
    path = models.CharField(max_length=2000, blank=True)
    path_verbose = models.CharField(max_length=2000, blank=True)
    field = models.CharField(max_length=2000)
    field_verbose = models.CharField(max_length=2000)
    position = models.PositiveSmallIntegerField(blank=True, null=True)

    class Meta:
        abstract = True
        ordering = ['position']

    @property
    def field_type(self):
        return self.report.get_field_type(self.field, self.path)

    @property
    def field_key(self):
        """ This key can be passed to a Django ORM values_list """
        return self.path + self.field

    @property
    def choices(self):
        if self.pk:
            model = get_model_from_path_string(
                self.report.root_model.model_class(), self.path)
            return self.get_choices(model, self.field)


class DisplayField(AbstractField):
    """ A display field to show in a report. Always belongs to a Report
    """
    name = models.CharField(max_length=2000)
    sort = models.IntegerField(blank=True, null=True)
    sort_reverse = models.BooleanField(verbose_name="Reverse", default=False)
    width = models.IntegerField(default=15)
    aggregate = models.CharField(
        max_length=5,
        choices=(
            ('Sum', 'Sum'),
            ('Count', 'Count'),
            ('Avg', 'Avg'),
            ('Max', 'Max'),
            ('Min', 'Min'),
        ),
        blank=True
    )
    total = models.BooleanField(default=False)
    group = models.BooleanField(default=False)
    display_format = models.ForeignKey(Format, blank=True, null=True)

    def get_choices(self, model, field_name):
        try:
            model_field = model._meta.get_field_by_name(field_name)[0]
        except:
            model_field = None
        if model_field and model_field.choices:
            return ((model_field.get_prep_value(key), val) for key, val in model_field.choices)

    @property
    def choices_dict(self):
        choices = self.choices
        choices_dict = {}
        if choices:
            for choice in choices:
                choices_dict.update({choice[0]: choice[1]})
        return choices_dict

    def __str__(self):
        return self.name


class FilterField(AbstractField):
    """
    A filter model used to filter DisplayFields using Django ORM
    filter arguments or custom filter values.
    """

    filter_type = models.CharField(
        max_length=20,
        choices=(
            ('exact', 'Equals'),
            ('iexact', 'Equals (case-insensitive)'),
            ('contains', 'Contains'),
            ('icontains', 'Contains (case-insensitive)'),
            ('in', 'in (comma seperated 1,2,3)'),
            ('gt', 'Greater than'),
            ('gte', 'Greater than equals'),
            ('lt', 'Less than'),
            ('lte', 'Less than equals'),
            ('startswith', 'Starts with'),
            ('istartswith', 'Starts with (case-insensitive)'),
            ('endswith', 'Ends with'),
            ('iendswith', 'Ends with  (case-insensitive)'),
            ('range', 'range'),
            ('relative_range', 'relative_range'),
            ('week_day', 'Week day'),
            ('isnull', 'Is null'),
            ('regex', 'Regular Expression'),
            ('iregex', 'Reg. Exp. (case-insensitive)'),
            ('max', 'Max (annotation-filter)'),
            ('min', 'Min (annotation-filter)'),
        ),
        blank=True,
        default='icontains',
    )
    filter_delta = models.BigIntegerField(null=True, blank=True)
    filter_value = models.CharField(max_length=2000)
    filter_value2 = models.CharField(max_length=2000, blank=True)
    exclude = models.BooleanField(default=False)

    def clean(self):
        dt_types = ['DateField', 'DateTimeField', 'TimeField']

        if self.filter_type == 'range' and self.filter_value2 in [None, '']:
            raise ValidationError('Range filters must have two values')

        # Raise error if 'relative_range' filter is applied to a non-supported
        # field type
        if self.filter_type == 'relative_range' and self.field_type not in dt_types:
            raise ValidationError(
                'Relative Range filtering is only currently supported for'
                ' the following field types: {}.'.format(dt_types))

        # Check for required relative range filter_delta
        if self.filter_type == 'relative_range' and self.filter_delta is null:
            raise ValidationError(
                'Relative Range filters must have value and delta inputs.')

        if self.filter_type in ('max', 'min'):
            # These filter types ignore their value.
            pass

        # clean DateTimeField inputs
        elif self.field_type == 'DateField' and self.filter_type != 'isnull':
            self.filter_value = str(self.parse_datetime_fields(self.filter_value))

        return super(FilterField, self).clean()

    def parse_datetime_fields(self, dt_type):
        """Clean and parse datetime filter_value inputs."""
        date_value = parser.parse(dt_type)
        date_form = forms.DateTimeField()

        if self.field_type == 'DateField':
            date_value = date_value.date()
            date_form = forms.DateField()
        if self.field_type == 'TimeField':
            date_value = date_value.time()
            date_form = forms.TimeField()

        return date_form.clean(date_value)

    def get_choices(self, model, field_name):
        try:
            model_field = model._meta.get_field_by_name(field_name)[0]
        except:
            model_field = None
        if model_field and model_field.choices:
            return model_field.choices

    def filter_property(self, value):
        """ Determine if passed value should be filtered or not """
        filter_field = self
        filter_type = filter_field.filter_type
        filter_value = filter_field.filter_value
        filtered = True
        WEEKDAY_INTS = {
            'monday': 0,
            'tuesday': 1,
            'wednesday': 2,
            'thursday': 3,
            'friday': 4,
            'saturday': 5,
            'sunday': 6,
        }
        if filter_type == 'exact' and str(value) == filter_value:
            filtered = False
        if filter_type == 'iexact' and str(value).lower() == str(filter_value).lower():
            filtered = False
        if filter_type == 'contains' and filter_value in value:
            filtered = False
        if filter_type == 'icontains' and str(filter_value).lower() in str(value).lower():
            filtered = False
        if filter_type == 'in' and value in filter_value:
            filtered = False
        # convert dates and datetimes to timestamps in order to compare digits and date/times the same
        if isinstance(value, datetime.datetime) or isinstance(value, datetime.date):
            value = str(time.mktime(value.timetuple()))
            try:
                filter_value_dt = parser.parse(filter_value)
                filter_value = str(time.mktime(filter_value_dt.timetuple()))
            except ValueError:
                pass
        if filter_type == 'gt' and Decimal(value) > Decimal(filter_value):
            filtered = False
        if filter_type == 'gte' and Decimal(value) >= Decimal(filter_value):
            filtered = False
        if filter_type == 'lt' and Decimal(value) < Decimal(filter_value):
            filtered = False
        if filter_type == 'lte' and Decimal(value) <= Decimal(filter_value):
            filtered = False
        if filter_type == 'startswith' and str(value).startswith(str(filter_value)):
            filtered = False
        if filter_type == 'istartswith' and str(value).lower().startswith(str(filter_value)):
            filtered = False
        if filter_type == 'endswith' and str(value).endswith(str(filter_value)):
            filtered = False
        if filter_type == 'iendswith' and str(value).lower().endswith(str(filter_value)):
            filtered = False
        if filter_type == 'range' and value in [int(x) for x in filter_value]:
            filtered = False
        if filter_type == 'week_day' and WEEKDAY_INTS.get(str(filter_value).lower()) == value.weekday:
            filtered = False
        if filter_type == 'isnull' and value is None:
            filtered = False
        if filter_type == 'regex' and re.search(filter_value, value):
            filtered = False
        if filter_type == 'iregex' and re.search(filter_value, value, re.I):
            filtered = False

        if filter_field.exclude:
            return not filtered
        return filtered

    def get_relative_range(self):
        """
        Generate a 'filterable' range from the current date and filter delta.
        Ex.
            With:
                self.filter_type = 'relative_range'
                self.filter_delta = -60 * 60 * 24 * 2 (i.e. -2 days)
            Return:
                # a 'negative' two day range from filter_value
                ["2017-01-01", "2017-01-03"]
        """
        day = 60 * 60 * 24

        if self.field_type == 'DateField':
            if abs(self.filter_delta) < day:
                raise ValidationError(
                    'DateField delta must be at least 1 day.')
            first = datetime.date.today()
            second = first + datetime.timedelta(seconds=self.filter_delta)
            output_range = sorted([first, second])
            output = [date.strftime("%Y-%m-%d") for date in output_range]

        elif self.field_type == 'DateTimeField':
            if abs(self.filter_delta) < day:
                raise ValidationError(
                    'DateTimeField delta must be at least 1 day.')
            first = datetime.datetime.today()
            second = first + datetime.timedelta(seconds=self.filter_delta)
            output_range = sorted([first, second])
            output = [date.strftime("%Y-%m-%d %H:%M:%S") for date in output_range]

        elif self.field_type == 'TimeField':
            day_const = datetime.datetime(1, 1, 1)
            first = datetime.datetime.now().time()
            delta = datetime.timedelta(seconds=self.filter_delta)
            diff = datetime.datetime.combine(day_const, first) + delta
            second = diff.time()
            output_range = sorted([first, second])
            output = [date.strftime("%H:%M") for date in output_range]

        return output

    @property
    def field_type(self):
        return self.report.get_field_type(self.field, self.path)

    @property
    def choices(self):
        if self.pk:
            model = get_model_from_path_string(
                self.report.root_model.model_class(), self.path)
            return self.get_choices(model, self.field)

    def __str__(self):
        return self.field



from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import reverse
from django.db.models.query import QuerySet
from django.test import TestCase
from django.test.utils import override_settings
from ..models import (
    Report, DisplayField, FilterField, Format, get_allowed_models,
    get_limit_choices_to_callable)
from report_builder_demo.demo_models.models import (
    Bar, Place, Restaurant, Waiter, Person, Child)
from django.conf import settings
from rest_framework.test import APIClient
import time
import csv
import unittest
from io import StringIO
from freezegun import freeze_time
from datetime import date, datetime, timedelta, time as dtime

from django.contrib.auth import get_user_model
User = get_user_model()


def find_duplicates_in_contexttype():
    find_duplicates = {}
    duplicates = []
    for i in get_allowed_models():
        model_name = str(i.model)
        if model_name in find_duplicates:
            find_duplicates[model_name] += 1
            duplicates.append(model_name)
        else:
            find_duplicates[model_name] = 1
    return duplicates


class ReportBuilderTests(TestCase):
    def setUp(self):
        user = User.objects.get_or_create(username='testy')[0]
        user.is_staff = True
        user.is_superuser = True
        user.set_password('pass')
        user.save()
        self.client = APIClient()
        self.client.login(username='testy', password='pass')

    def test_get_allowed_models_for_include(self):
        pre_include_duplicates = find_duplicates_in_contexttype()
        settings.REPORT_BUILDER_INCLUDE = (
            'demo_models.bar',
            'demo_models.foo',
            'demo_models.place',
        )
        post_include_duplicates = find_duplicates_in_contexttype()
        settings.REPORT_BUILDER_INCLUDE = None
        self.assertEqual(pre_include_duplicates, ['bar'])
        self.assertEqual(post_include_duplicates, [])

    def test_get_allowed_models_for_exclude(self):
        pre_exclude_duplicates = find_duplicates_in_contexttype()
        settings.REPORT_BUILDER_EXCLUDE = (
            'demo_second_app.bar',
        )
        post_exclude_duplicates = find_duplicates_in_contexttype()
        settings.REPORT_BUILDER_EXCLUDE = None
        self.assertEqual(pre_exclude_duplicates, ['bar'])
        self.assertEqual(post_exclude_duplicates, [])

    def test_get_allowed_models_lookup_dict(self):
        settings.REPORT_BUILDER_INCLUDE = (
            'demo_models.bar',
            'demo_models.foo',
            'demo_models.place',
        )
        models = get_allowed_models()
        lookup_dict = get_limit_choices_to_callable()

        self.assertTrue(callable(get_limit_choices_to_callable))
        self.assertTrue(isinstance(lookup_dict['pk__in'], QuerySet))
        self.assertQuerysetEqual(lookup_dict['pk__in'], map(repr, models), ordered=False)
    
    def test_report_builder_reports(self):
        url = '/report_builder/api/reports/'
        no_auth_client = APIClient()

        # A non authenticated client cannot view
        response = no_auth_client.get(url)
        self.assertEqual(response.status_code, 403)

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_report_builder_fields(self):
        ct = ContentType.objects.get(model="foo", app_label="demo_models")
        response = self.client.post(
            '/report_builder/api/fields/',
            {"model": ct.id, "path": "", "path_verbose": "", "field": ""})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'char_field')
        self.assertNotContains(response, 'char_field2')

    def test_report_builder_fields_from_related(self):
        ct = ContentType.objects.get(model="place", app_label="demo_models")
        response = self.client.post(
            '/report_builder/api/fields/',
            {"model": ct.id,
             "path": "",
             "path_verbose": "",
             "field": "restaurant"})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'pizza')

    def test_report_builder_fields_from_related_fields(self):
        ct = ContentType.objects.get(model="place", app_label="demo_models")
        response = self.client.post(
            '/report_builder/api/related_fields/',
            {"model": ct.id,
             "path": "",
             "path_verbose": "",
             "field": "restaurant"})
        self.assertContains(response, '"parent_model_name"')
        self.assertContains(response, '"parent_model_app_label"')
        self.assertContains(response, '"included_model"')
        self.assertEqual(response.status_code, 200)

    def test_report_builder_exclude(self):
        ct = ContentType.objects.get(model="fooexclude", app_label="demo_models")
        response = self.client.post(
            '/report_builder/api/fields/',
            {"model": ct.id, "path": "", "path_verbose": "", "field": ""})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'char_field')
        self.assertNotContains(response, 'char_field2')

    def test_report_builder_extra(self):
        ct = ContentType.objects.get(model="bar", app_label="demo_models")
        response = self.client.post(
            '/report_builder/api/fields/',
            {"model": ct.id, "path": "", "path_verbose": "", "field": ""}
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'char_field')
        self.assertContains(response, 'i_want_char_field')
        self.assertContains(response, 'i_need_char_field')

    def test_report_builder_is_default(self):
        ct = ContentType.objects.get(model="bar", app_label="demo_models")
        response = self.client.post(
            '/report_builder/api/fields/',
            {"model": ct.id, "path": "", "path_verbose": "", "field": ""})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'is_default')

    def test_report_builder_choices(self):
        ct = ContentType.objects.get(model="bar", app_label="demo_models")
        response = self.client.post(
            '/report_builder/api/fields/',
            {"model": ct.id, "path": "", "path_verbose": "", "field": ""})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'field_choices')
        self.assertContains(response, '[["CH","CHECK"],["MA","CHECKMATE"]]')

    def test_report_builder_can_filter(self):
        ct = ContentType.objects.get(model="bar", app_label="demo_models")
        response = self.client.post(
            '/report_builder/api/fields/',
            {"model": ct.id, "path": "", "path_verbose": "", "field": ""})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'can_filter')
        for field in response.data:
            if field['name'] == 'char_field':
                self.assertEqual(field['can_filter'], True)
            else:
                self.assertEqual(field['can_filter'], False)
        # Now confirm not setting filter makes it always true
        ct = ContentType.objects.get(model="foo", app_label="demo_models")
        response = self.client.post(
            '/report_builder/api/fields/',
            {"model": ct.id, "path": "", "path_verbose": "", "field": ""})
        for field in response.data:
            self.assertEqual(field['can_filter'], True)

class ReportTests(TestCase):

    def setUp(self):
        user = User.objects.get_or_create(username='testy')[0]
        user.is_staff = True
        user.is_superuser = True
        user.set_password('pass')
        user.save()
        self.client = APIClient()
        self.client.login(username='testy', password='pass')
        ct = ContentType.objects.get(model="bar", app_label="demo_models")
        self.report = Report.objects.create(root_model=ct, name="A")
        self.bar = Bar.objects.create(char_field="wooo")
        self.generate_url = reverse('generate_report', args=[self.report.id])
        # time increments in seconds
        self.min = 60
        self.hour = self.min * self.min
        self.day = self.hour * 24

    def test_property_position(self):
        bar = self.bar

        DisplayField.objects.create(
            report=self.report,
            field="i_want_char_field",
            field_verbose="stuff",
            position=1,
        )
        DisplayField.objects.create(
            report=self.report,
            field="i_need_char_field",
            field_verbose="stuff",
            position=2,
        )
        report_list = self.report.report_to_list(self.report.get_query())
        self.assertEqual(
            report_list[0],
            [bar.i_want_char_field, bar.i_need_char_field]
        )

    def test_property_and_field_position(self):
        bar = self.bar

        DisplayField.objects.create(
            report=self.report,
            field="char_field",
            field_verbose="stuff",
        )
        DisplayField.objects.create(
            report=self.report,
            field="i_want_char_field",
            field_verbose="stuff",
        )
        DisplayField.objects.create(
            report=self.report,
            field="i_need_char_field",
            field_verbose="stuff",
        )
        DisplayField.objects.create(
            report=self.report,
            field="char_field",
            field_verbose="stuff",
        )

        report_list = self.report.report_to_list(self.report.get_query())
        self.assertEqual(
            report_list[0],
            [bar.char_field, bar.i_want_char_field, bar.i_need_char_field, bar.char_field]
        )

    def test_property_display(self):
        DisplayField.objects.create(
            report=self.report,
            field="i_want_char_field",
            field_verbose="stuff",
        )
        response = self.client.get(self.generate_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'lol no')

    def test_cached_property_display(self):
        DisplayField.objects.create(
            report=self.report,
            field="i_need_char_field",
            field_verbose="stuff",
        )
        response = self.client.get(self.generate_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'lol yes')

    def test_error_display_field(self):
        DisplayField.objects.create(
            report=self.report,
            field="i_do_not_exist",
            field_verbose="stuff",
            position=0,
        )
        DisplayField.objects.create(
            report=self.report,
            field="i_want_char_field",
            field_verbose="stuff",
            position=1,
        )
        DisplayField.objects.create(
            report=self.report,
            field="i_need_char_field",
            field_verbose="stuff",
            position=2,
        )
        response = self.client.get(self.generate_url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'lol no')
        self.assertContains(response, 'lol yes')
        self.assertNotContains(response, 'i_do_not_exist')

    def test_filter_property_contains(self):
        DisplayField.objects.create(
            report=self.report,
            field="i_want_char_field",
            field_verbose="stuff",
        )
        filter_field = FilterField.objects.create(
            report=self.report,
            field="i_want_char_field",
            field_verbose="stuff",
            filter_type='contains',
            filter_value='lol no',
        )
        response = self.client.get(self.generate_url)
        self.assertContains(response, 'lol no')
        filter_field.filter_value = 'It does not contain this'
        filter_field.save()
        response = self.client.get(self.generate_url)
        self.assertNotContains(response, 'lol no')

    def test_filter_cached_property(self):
        DisplayField.objects.create(
            report=self.report,
            field="i_need_char_field",
            field_verbose="stuff",
        )
        filter_field = FilterField.objects.create(
            report=self.report,
            field="i_need_char_field",
            field_verbose="stuff",
            filter_type='contains',
            filter_value='lol yes',
        )
        response = self.client.get(self.generate_url)
        self.assertContains(response, 'lol yes')
        filter_field.filter_value = 'It does not contain this'
        filter_field.save()
        response = self.client.get(self.generate_url)
        self.assertNotContains(response, 'lol yes')

    def make_lots_of_foos(self):
        for x in range(500):
            bar = Bar.objects.create(char_field="wooo" + str(x))
            bar.foos.create(char_field="a")

    def test_performance(self):
        """ Test getting a report with ORM and property fields.
        Provides baseline on performance testing. """
        self.make_lots_of_foos()
        DisplayField.objects.create(
            report=self.report,
            field="i_want_char_field",
            field_verbose="stuff",
        )
        DisplayField.objects.create(
            report=self.report,
            field="char_field",
            field_verbose="stuff",
        )
        start = time.time()
        response = self.client.get(self.generate_url)
        run_time = time.time() - start
        print('report builder report time is {}'.format(run_time))
        self.assertEqual(response.status_code, 200)
        self.assertLess(run_time, 1.0)

    def test_performance_filter(self):
        """ Test getting a report with ORM and property fields.
        Provides baseline on performance testing. """
        self.make_lots_of_foos()
        DisplayField.objects.create(
            report=self.report,
            field="i_want_char_field",
            field_verbose="stuff",
        )
        DisplayField.objects.create(
            report=self.report,
            field="i_need_char_field",
            field_verbose="stuff",
        )
        DisplayField.objects.create(
            report=self.report,
            field="char_field",
            field_verbose="stuff",
        )
        FilterField.objects.create(
            report=self.report,
            field="i_want_char_field",
            field_verbose="stuff",
            filter_type='contains',
            filter_value='I am custom',
        )
        start = time.time()
        response = self.client.get(self.generate_url)
        run_time = time.time() - start
        print('report builder report time is {}'.format(run_time))
        self.assertEqual(response.status_code, 200)
        self.assertLess(run_time, 1.0)

    def make_tiny_town(self):
        data = [
            ('Golden Gate Bridge', '123 Bridge St, SF, CA', True, True, 4),
            ('Alcatraz Island', '0 Dead Mans Row, SF Bay, CA', True, False, 2),
            ("Fisherman's Warf", 'Pier 39, San Francisco, CA', False, True, 3),
            ('AT&T Park', '3 World Series Way, SF, CA', False, False, 5),
            ('MOMA', '456 Art Avenue, San Francisco, CA', True, True, 8),
        ]

        total = 0

        for row in data:
            place = Place.objects.create(name=row[0], address=row[1])
            restaurant = Restaurant.objects.create(
                place=place, serves_hot_dogs=row[2], serves_pizza=row[3]
            )

            for count in range(row[4]):
                days = None if not (total % 3 | total % 2) else total % 3

                Waiter.objects.create(
                    restaurant=restaurant,
                    name=str(total),
                    days_worked=days,
                )

                total += 1

    def test_total_accounting(self):
        """ Test accounting total fields.
        Nullable fields should be totalled as 0.
        """
        self.make_tiny_town()

        model = ContentType.objects.get(model='waiter', app_label="demo_models")
        report = Report.objects.create(root_model=model, name='Waiter Days Worked')

        DisplayField.objects.create(
            report=report,
            field='name',
            field_verbose='name_verbose',
            total=True,
            position=0,
        )
        DisplayField.objects.create(
            report=report,
            field='days_worked',
            field_verbose='days_worked_verbose',
            total=True,
            position=1,
        )

        generate_url = reverse('generate_report', args=[report.id])
        response = self.client.get(generate_url)

        self.assertContains(response, '["TOTALS",""],[22.0,21.0]')

    @freeze_time("2017-11-01 12:00:00")
    def make_people(self):
        """
        Make mock data from the Person and Child demo models.
        Follows the format:
            Person()
                first_name      (CharField)
                last_name       (CharField)
                last_modifed    (DateField)
                birth_date      (DateTimeField)
                hammer_time     (TimeField)
            Child()
                parent          (ForeignKey(Person))
                first_name      (CharField)
                last_name       (CharField)
                age             (IntegerField)
                color           (CharField)
        """
        people = (
            ('John', 'Doe', (
                ('Will', 'Doe', 5, 'R'),
                ('James', 'Doe', 8, ''),
                ('Robert', 'Doe', 3, 'G'),
            ), date.today() - timedelta(seconds=self.day * 5),
                datetime.today() - timedelta(seconds=self.day),
                dtime(hour=12)),
            ('Maria', 'Smith', (
                ('Susan', 'Smith', 1, 'Y'),
                ('Karen', 'Smith', 4, 'B'),
            ), date.today() - timedelta(seconds=self.day * 10),
                datetime.today() - timedelta(seconds=self.day * 30),
                dtime(hour=16)),
            ('Donald', 'King', (
                ('Charles', 'King', None, ''),
                ('Helen', 'King', 7, 'G'),
                ('Mark', 'King', 2, 'Y'),
                ('Karen', 'King', 4, 'R'),
                ('Larry', 'King', 5, 'R'),
                ('Lisa', 'King', 3, 'R'),
            ), datetime.today() - timedelta(seconds=self.day * 15),
                datetime.today() - timedelta(seconds=self.day * 60),
                dtime(hour=20)),
            ('Paul', 'Nelson', (),
                date.today() - timedelta(seconds=self.day * 20),
                datetime.today() - timedelta(seconds=self.day * 90),
                dtime(hour=22)),
        )
        for first, last, cn, lm, bd, ht in people:
            person = Person(
                first_name=first,
                last_name=last,
                last_modifed=lm,  # DateField
                birth_date=bd,  # DateTimeField
                hammer_time=ht)  # TimeField
            person.save()
            for child_first, child_last, age, color in cn:
                child = Child(
                    parent=person, first_name=child_first, last_name=child_last,
                    age=age, color=color
                )
                child.save()

    @freeze_time("2017-11-01 12:00:00")
    def make_people_report(self):
        """
        Make a mock report using the People demo model.

        Creates mock People instances and returns a report including the 
        following DisplayFields:
            first_name      (CharField)
            last_modifed    (DateField)
            birth_date      (DateTimeField)
            hammer_time     (TimeField)
        """ 
        self.make_people()
        model = ContentType.objects.get(model="person",
                                        app_label="demo_models")
        people_report = Report.objects.create(
            root_model=model,
            name="A report of people")

        DisplayField.objects.create(
            report=people_report,
            field='first_name',
            field_verbose='First Name',
            sort=4,
            position=0,
        )

        DisplayField.objects.create(
            report=people_report,
            field='last_modifed',
            field_verbose='Last Modifed',
            sort=2,
            position=2,
        )

        DisplayField.objects.create(
            report=people_report,
            field='birth_date',
            field_verbose='Birth Date',
            sort=1,
            position=3,
        )

        DisplayField.objects.create(
            report=people_report,
            field='hammer_time',
            field_verbose='Hammer Time',
            sort=1,
            position=3,
        )
        return people_report

    @freeze_time("2017-11-01 12:00:00")
    def test_filter_datetime_lte_filter(self):
        """
        Test filtering 'DateTime' field types using simple ORM filtering
        (i.e. filter_type='lte')
        """
        people_report = self.make_people_report()

        # DateField
        ff = FilterField.objects.create(
            report=people_report,
            field='last_modifed', 
            filter_type='lte',
            filter_value=str(date.today() - timedelta(seconds=self.day * 10)),
        )

        generate_url = reverse('generate_report', args=[people_report.id])
        response = self.client.get(generate_url)
        # filter from 4 to 2 people
        self.assertEquals(len(response.data['data']), 3)

        # TimeField
        ff.field = 'hammer_time' 
        ff.filter_value = str(dtime(hour=13))
        ff.save()

        response = self.client.get(generate_url)
        self.assertEquals(len(response.data['data']), 1)

        # DateTimeField
        ff.field = 'birth_date' 
        ff.filter_value = str(datetime.today() - timedelta(seconds=self.day * 40))
        ff.save()

        response = self.client.get(generate_url)
        self.assertEquals(len(response.data['data']), 2)

    @freeze_time("2017-11-01 12:00:00")
    def test_filter_datetime_range(self):
        """
        Test filtering 'DateTime' field types using a range filter.
        
        Each FilterField accepts 2 values of the respective field type to 
        create the range.

        Ex. Filter a TimeField for 'user logins between 10am - 1pm':
            filter_type='range',
            filter_value("HH:MM") = "10:00"
            filter_value2("HH:MM") = "13:00"
        """
        people_report = self.make_people_report()

        # DateField
        ff = FilterField.objects.create(
            report=people_report,
            field='last_modifed', 
            filter_type='range',
            filter_value=str(date.today() - timedelta(seconds=self.day * 7)),
            filter_value2=str(date.today()),
        )

        generate_url = reverse('generate_report', args=[people_report.id])
        response = self.client.get(generate_url)
        self.assertEquals(len(response.data['data']), 1)

        # TimeField
        ff.field = 'hammer_time' 
        ff.filter_value = str(dtime(hour=10))
        ff.filter_value2 = str(dtime(hour=13))
        ff.save()

        response = self.client.get(generate_url)
        self.assertEquals(len(response.data['data']), 1)

        # DateTimeField
        ff.field = 'birth_date' 
        ff.filter_value = str(datetime.now() - timedelta(seconds=self.day * 50))
        ff.filter_value2 = str(datetime.now() - timedelta(seconds=self.day * 70))
        ff.save()

        response = self.client.get(generate_url)
        self.assertEquals(len(response.data['data']), 1)

    @freeze_time("2017-11-01 12:00:00")
    def test_filter_datetime_relative_range(self):
        """
        Test filtering 'DateTime' field types using a relative range filter.

        Each FilterField accepts a delta value (in seconds) which represents 
        the range (positive or negative) off of the current date. 

        Ex. Filter a TimeField for 'user logins in the last 3 hours':
            filter_type='relative_range',
            filter_delta = -60 * 60 * 3 (i.e. 3 hours)
        """
        people_report = self.make_people_report()

        # DateField w/ full day
        ff = FilterField.objects.create(
            report=people_report,
            field='last_modifed',
            filter_type='relative_range',
            filter_delta=self.day * -7,
        )

        generate_url = reverse('generate_report', args=[people_report.id])
        response = self.client.get(generate_url)
        self.assertEquals(len(response.data['data']), 1)

        # DateField w/ partial day
        ff.filter_delta = self.day * -7 + 5 
        ff.save()

        response = self.client.get(generate_url)
        self.assertEquals(len(response.data['data']), 1)

        # TimeField w/ hour delta
        ff.field = 'hammer_time'
        ff.filter_delta = self.hour * 5
        ff.save()

        response = self.client.get(generate_url)
        self.assertEquals(len(response.data['data']), 2)

        # TimeField w/ sec delta
        ff.filter_delta = -5
        ff.save()

        response = self.client.get(generate_url)
        self.assertEquals(len(response.data['data']), 1)

        # DateTimeField w/ full day
        ff.field = 'birth_date'
        ff.filter_delta = self.day * -30
        ff.save()

        response = self.client.get(generate_url)
        self.assertEquals(len(response.data['data']), 2)

        # # DateTimeField w/ partial day
        ff.filter_delta = (self.day + self.hour) * -1
        ff.save()

        response = self.client.get(generate_url)
        self.assertEquals(len(response.data['data']), 1)

    def test_filter_datefield_relative_range_over_time(self):
        """
        Test filtering DateField types using a relative range filter
        over time. 
        """
        people_report = self.make_people_report()
        generate_url = reverse('generate_report', args=[people_report.id])

        initial_today = datetime(2017, 11, 1, 12)
        ten_days_later = datetime(2017, 11, 10, 12)

        # DateField with login 'today'
        with freeze_time(initial_today) as frozen_today:
            FilterField.objects.create(
                report=people_report,
                field='last_modifed',
                filter_type='relative_range',
                filter_delta=self.day * -16,
            )
            response = self.client.get(generate_url)
            self.assertEquals(len(response.data['data']), 3)

            # login again 10 days later
            frozen_today.move_to(ten_days_later)
            response = self.client.get(generate_url)
            self.assertEquals(len(response.data['data']), 1)

    def test_filter_timefield_relative_range_over_time(self):
        """
        Test filtering TimeField field types using a relative range filter
        over time. 
        """
        
        people_report = self.make_people_report()
        generate_url = reverse('generate_report', args=[people_report.id])
        
        initial_today = datetime(2017, 11, 1, 12)
        four_hours_later_today = datetime(2017, 11, 1, 16)
        
        # TimeField with login 'now'
        with freeze_time(initial_today) as frozen_today:
            FilterField.objects.create(
                report=people_report,
                field='hammer_time',
                filter_type='relative_range',
                filter_delta=self.hour * -10,
            )
            response = self.client.get(generate_url)
            self.assertEquals(len(response.data['data']), 1)

            # login 4 hours later
            frozen_today.move_to(four_hours_later_today)
            response = self.client.get(generate_url)
            self.assertEquals(len(response.data['data']), 2)

    def test_filter_datetimefield_relative_range_over_time(self):
        """
        Test filtering DateTimeField field types using a relative range filter
        over time. 
        """
        people_report = self.make_people_report()
        generate_url = reverse('generate_report', args=[people_report.id])
        
        initial_today = datetime(2017, 10, 1, 12)
        one_month_later = datetime(2017, 11, 1, 12)

        # DateTimeField with login today
        with freeze_time(initial_today) as frozen_today:
            FilterField.objects.create(
                report=people_report,
                field='birth_date',
                filter_type='relative_range',
                filter_delta=self.day * -30
            )

            response = self.client.get(generate_url)
            self.assertEquals(len(response.data['data']), 1)

            # # login one month later
            frozen_today.move_to(one_month_later)
            response = self.client.get(generate_url)
            self.assertEquals(len(response.data['data']), 2)

    def test_groupby_id(self):
        self.make_people()

        model = ContentType.objects.get(model='child', app_label="demo_models")
        report = Report.objects.create(root_model=model, name='# of Kids / Person Id')

        DisplayField.objects.create(
            report=report,
            path='parent__',
            field='id',
            field_verbose="Parent's Id",
            sort=3,
            position=0,
            group=True,
        )

        DisplayField.objects.create(
            report=report,
            path='parent__',
            field='first_name',
            field_verbose="Parent's First name",
            sort=2,
            sort_reverse=True,
            position=1,
        )

        DisplayField.objects.create(
            report=report,
            path='parent__',
            field='last_name',
            field_verbose="Parent's Last name",
            sort=1,
            position=2,
        )

        DisplayField.objects.create(
            report=report,
            field='id',
            field_verbose='Child Count',
            sort=4,
            position=3,
            aggregate='Count',
            total=True,
        )

        generate_url = reverse('generate_report', args=[report.id])
        response = self.client.get(generate_url)

        data = '"data":[[1,"John","Doe",3],[3,"Donald","King",6],[2,"Maria","Smith",2],["TOTALS","","",""],["","","",11.0]]'

        self.assertContains(response, data)

    def test_groupby_name(self):
        self.make_people()

        model = ContentType.objects.get(model='child', app_label="demo_models")
        report = Report.objects.create(root_model=model, name='# of Kids / Person Name')

        DisplayField.objects.create(
            report=report,
            path='parent__',
            field='id',
            field_verbose="Parent's Id",
            sort=3,
            position=0,
        )

        DisplayField.objects.create(
            report=report,
            path='parent__',
            field='first_name',
            field_verbose="Parent's First name",
            sort=2,
            sort_reverse=True,
            position=1,
            group=True,
        )

        DisplayField.objects.create(
            report=report,
            path='parent__',
            field='last_name',
            field_verbose="Parent's Last name",
            sort=1,
            position=2,
            group=True,
            total=True,
        )

        DisplayField.objects.create(
            report=report,
            field='id',
            field_verbose='Child Count',
            sort=4,
            position=3,
            aggregate='Count',
            total=True,
        )

        generate_url = reverse('generate_report', args=[report.id])
        response = self.client.get(generate_url)

        data = '"data":[[1,"John","Doe",3],[3,"Donald","King",6],[2,"Maria","Smith",2],["TOTALS","","",""],["","",3.0,11.0]]'

        self.assertContains(response, data)

    def test_aggregates(self):
        self.make_people()

        model = ContentType.objects.get(model='person', app_label="demo_models")
        report = Report.objects.create(root_model=model, name='Kid Stats')

        DisplayField.objects.create(
            report=report,
            field='first_name',
            field_verbose='First Name',
            sort=1,
            position=0,
        )

        DisplayField.objects.create(
            report=report,
            field='last_name',
            field_verbose='Last Name',
            sort=2,
            position=1,
        )

        DisplayField.objects.create(
            report=report,
            path='children__',
            field='age',
            field_verbose='Oldest Age',
            sort=3,
            position=2,
            aggregate='Max',
        )

        DisplayField.objects.create(
            report=report,
            path='children__',
            field='age',
            field_verbose='Youngest Age',
            sort=4,
            position=3,
            aggregate='Min',
        )

        DisplayField.objects.create(
            report=report,
            path='children__',
            field='age',
            field_verbose='Average Age',
            sort=5,
            position=4,
            aggregate='Avg',
        )

        DisplayField.objects.create(
            report=report,
            path='children__',
            field='age',
            field_verbose='Parenting Years',
            sort=6,
            position=5,
            aggregate='Sum',
        )

        generate_url = reverse('generate_report', args=[report.id])
        response = self.client.get(generate_url)

        data = '"data":[["Donald","King",7,2,4.2,21],["John","Doe",8,3,5.333333333333333,16],["Maria","Smith",4,1,2.5,5],["Paul","Nelson",null,null,null,null]]'

        self.assertContains(response, data)

    @unittest.skip('Broken in 1.10')
    def test_choices_and_sort_null(self):
        self.make_people()

        model = ContentType.objects.get(model='person', app_label="demo_models")
        report = Report.objects.create(root_model=model, name='Kid Data')

        DisplayField.objects.create(
            report=report,
            field='first_name',
            field_verbose='First Name',
            sort=4,
            position=0,
        )

        DisplayField.objects.create(
            report=report,
            field='last_name',
            field_verbose='Last Name',
            sort=3,
            position=1,
        )

        DisplayField.objects.create(
            report=report,
            path='children__',
            field='age',
            field_verbose='Child Age',
            sort=2,
            position=2,
        )

        DisplayField.objects.create(
            report=report,
            path='children__',
            field='color',
            field_verbose='Child Color',
            sort=1,
            position=3,
        )

        FilterField.objects.create(
            report=report,
            field='first_name',
            filter_type='exact',
            filter_value='Donald',
        )

        generate_url = reverse('generate_report', args=[report.id])
        response = self.client.get(generate_url)

        data = '"data":[["Donald","King",null,""],["Donald","King",7,"Green"],["Donald","King",3,"Red"],["Donald","King",4,"Red"],["Donald","King",5,"Red"],["Donald","King",2,"Y"]]'

        self.assertContains(response, data)

    def test_formatter(self):
        self.make_people()

        model = ContentType.objects.get(model='child', app_label="demo_models")
        report = Report.objects.create(root_model=model, name='Children')

        DisplayField.objects.create(
            report=report,
            field='first_name',
            field_verbose='First Name',
            sort=1,
            position=0,
        )

        DisplayField.objects.create(
            report=report,
            field='last_name',
            field_verbose='Last Name',
            sort=2,
            position=1,
        )

        years_old = Format(name='years old', string='{} years old')
        years_old.save()

        DisplayField.objects.create(
            report=report,
            field='age',
            field_verbose='Child Age',
            sort=3,
            position=2,
            total=True,
            display_format=years_old,
        )

        FilterField.objects.create(
            report=report,
            path='parent__',
            field='first_name',
            filter_type='exact',
            filter_value='Maria',
        )

        generate_url = reverse('generate_report', args=[report.id])
        response = self.client.get(generate_url)

        data = '"data":[["Karen","Smith","4 years old"],["Susan","Smith","1 years old"],["TOTALS","",""],["","","5 years old"]]'

        self.assertContains(response, data)

    def test_csv(self):
        self.make_people()

        model = ContentType.objects.get(model='child', app_label="demo_models")
        report = Report.objects.create(root_model=model, name='Children')

        DisplayField.objects.create(
            report=report,
            field='first_name',
            field_verbose='First Name',
            sort=1,
            position=0,
        )

        DisplayField.objects.create(
            report=report,
            field='last_name',
            field_verbose='Last Name',
            sort=2,
            position=1,
        )

        years_old = Format(name='years old', string='{} years old')
        years_old.save()

        DisplayField.objects.create(
            report=report,
            field='age',
            field_verbose='Child Age',
            sort=3,
            position=2,
            total=True,
            display_format=years_old,
        )

        settings.REPORT_BUILDER_ASYNC_REPORT = False
        download_csv = reverse('report_download_file', args=[report.id, 'csv'])
        response = self.client.get(download_csv)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response._headers['content-type'][1], 'text/csv')
        csv_string = response._container[0]
        f = StringIO(csv_string.decode('UTF-8'))
        reader = csv.reader(f, delimiter=',')
        csv_list = list(reader)
        self.assertEqual(csv_list[1], ['Charles', 'King', 'None years old'])

    def test_admin(self):
        response = self.client.get('/admin/report_builder/report/')
        self.assertEqual(response.status_code, 200)

    @override_settings(REPORT_BUILDER_ASYNC_REPORT=False)
    def test_admin_sync(self):
        response = self.client.get('/admin/report_builder/report/')
        self.assertEqual(response.status_code, 200)

    def test_report_builder_related_fields(self):
        '''
        Test for Django 1.8 support via
        https://github.com/burke-software/django-report-builder/issues/144
        '''
        ct = ContentType.objects.get(model='place')
        response = self.client.post(
            '/report_builder/api/related_fields/', {
                'model': ct.id,
                'path': '',
                'field': 'restaurant'
            }
        )
        self.assertContains(response, '"field_name":"waiter"')
        self.assertContains(response, '"verbose_name":"waiter_set"')
        self.assertContains(response, '"path":"restaurant__"')

    def test_annotation_filter_min(self):
        """
        Similar to group-by queries, annotation-filters allow you to display only
        the Max or Min of a set of rows.

        Suppose we have two models, Person and Child. Each has first_name,
        last_name fields and Child has a foreign key to Person as parent. Child
        also has age and color fields. Already we can use the group-by
        mechanism to display each Person's count of children or the age of each
        person's oldest child.

        Unfortunately, getting the related items of an aggregate is difficult
        in Django and SQL. Suppose in the example above that we want to know
        the name and age of each person's oldest child. With annotation-filters
        this can be done.

        Annotation-filters are applied iteratively after normal queryset
        filtering. They work, as the name indicates, by applying Django's
        annotate and filter queryset methods. As they are self-referential,
        they do not accept a text input from the front-end.

        Annotation-filters are difficult to describe conceptually so I'll use
        an example. To display the name and age of each person's oldest child,
        we begin by constructing a report based on Child. We add display fields
        for the parent name, child name, and child age. Then for filter fields,
        we add the parent__children__age field. This lengthy path indicates
        that children are to be grouped by their parent and with the type Max,
        only the maximum age of each group should be included.
        """
        self.make_people()

        model = ContentType.objects.get(model='child')
        report = Report.objects.create(root_model=model, name="Person's Youngest Child")

        DisplayField.objects.create(
            report=report,
            path='parent__',
            field='first_name',
            field_verbose="Person First name",
            sort=2,
            position=0,
        )

        DisplayField.objects.create(
            report=report,
            path='parent__',
            field='last_name',
            field_verbose="Person Last name",
            sort=1,
            position=1,
        )

        DisplayField.objects.create(
            report=report,
            field='first_name',
            field_verbose='Child First Name',
            sort=4,
            position=2,
        )

        DisplayField.objects.create(
            report=report,
            field='last_name',
            field_verbose='Child Last Name',
            sort=3,
            position=3,
        )

        DisplayField.objects.create(
            report=report,
            field='age',
            field_verbose='Child Age',
            position=4,
        )

        FilterField.objects.create(
            report=report,
            path='parent__children__',
            field='age',
            filter_type='min',
            filter_value='True',
        )

        generate_url = reverse('generate_report', args=[report.id])
        response = self.client.get(generate_url)

        data = '"data":[["John","Doe","Robert","Doe",3],["Donald","King","Mark","King",2],["Maria","Smith","Susan","Smith",1]]'

        self.assertContains(response, data)

    def test_annotation_filter_max(self):
        """
        See test_annotation_filter_min for a description of annotation-filters.

        This test extends things further by adding the filter field
        parent__children__color with type Equals Red to get a list of the
        oldest children of each person whose favorite color is red.
        """
        self.make_people()

        model = ContentType.objects.get(model='child')
        report = Report.objects.create(root_model=model, name="Person's Oldest Child whose Favorite Color is Red")

        DisplayField.objects.create(
            report=report,
            path='parent__',
            field='first_name',
            field_verbose="Person First name",
            sort=2,
            position=0,
        )

        DisplayField.objects.create(
            report=report,
            path='parent__',
            field='last_name',
            field_verbose="Person Last name",
            sort=1,
            position=1,
        )

        DisplayField.objects.create(
            report=report,
            field='first_name',
            field_verbose='Child First Name',
            sort=4,
            position=2,
        )

        DisplayField.objects.create(
            report=report,
            field='last_name',
            field_verbose='Child Last Name',
            sort=3,
            position=3,
        )

        DisplayField.objects.create(
            report=report,
            field='age',
            field_verbose='Child Age',
            position=4,
        )

        FilterField.objects.create(
            report=report,
            path='parent__children__',
            field='color',
            filter_type='exact',
            filter_value='R',
        )

        FilterField.objects.create(
            report=report,
            path='parent__children__',
            field='age',
            filter_type='max',
            filter_value='True',
        )

        generate_url = reverse('generate_report', args=[report.id])
        response = self.client.get(generate_url)

        data = '"data":[["John","Doe","Will","Doe",5],["Donald","King","Larry","King",5]]'

        self.assertContains(response, data)

    def test_get_config(self):
        settings.REPORT_BUILDER_ASYNC_REPORT = True
        response = self.client.get('/report_builder/api/config/')
        self.assertContains(response,'"async_report": true')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('demo_models', '0003_auto_20150419_2110'),
    ]

    operations = [
        migrations.AddField(
            model_name='bar',
            name='date_field',
            field=models.DateField(null=True, blank=True),
        ),
    ]



import django
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core import mail
from django.core.urlresolvers import reverse
from django.test import TestCase
from .models import ScheduledReport
from report_builder.models import Report
from .tasks import report_builder_run_scheduled_report
from unittest import skipIf

User = get_user_model()


IS_D18 = False
if django.VERSION[0] is 1 and django.VERSION[1] is 8:
    IS_D18 = True


class ScheduledReportTests(TestCase):
    def test_scheduled_report(self):
        ct = ContentType.objects.get(model="bar", app_label="demo_models")
        report = Report.objects.create(root_model=ct, name="A")
        scheduled_report = ScheduledReport.objects.create(
            report=report,
            other_emails="test@example.com",
        )
        report_builder_run_scheduled_report(scheduled_report.id)
        scheduled_report.refresh_from_db()
        scheduled_report.report.refresh_from_db()
        self.assertIsNotNone(scheduled_report.last_run_at)
        self.assertIsNotNone(scheduled_report.report.report_file_creation)
        self.assertEqual(len(mail.outbox), 1)

    @skipIf(IS_D18, "Django 1.8 does not support force_login")
    def test_run_scheduled_report_view(self):
        ct = ContentType.objects.get(model="bar", app_label="demo_models")
        report = Report.objects.create(root_model=ct, name="A")
        scheduled_report = ScheduledReport.objects.create(
            report=report,
            other_emails="test@example.com",
        )

        url = reverse('run_scheduled_report', kwargs={'pk': scheduled_report.id})
        user = User.objects.create(username='testy', is_staff=True, is_superuser=True)
        self.client.force_login(user)
        res = self.client.get(url)
        self.assertEqual(res.status_code, 302)


class AdminViewTests(TestCase):
    """ Basic sanity check that admin views work """
    @skipIf(IS_D18, "Django 1.8 does not support force_login")
    def test_scheduled_report_admin(self):
        url = reverse('admin:report_builder_scheduled_scheduledreport_changelist')
        user = User.objects.create(username='testy', is_staff=True, is_superuser=True)
        self.client.force_login(user)
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)



from django.contrib import admin
from .models import (
    Foo, Bar, Place, Waiter, Restaurant, Account, Person, Child
)


@admin.register(Foo)
class FooAdmin(admin.ModelAdmin):
    pass


@admin.register(Bar)
class BarAdmin(admin.ModelAdmin):
    pass


admin.site.register(Account)
admin.site.register(Place)
admin.site.register(Waiter)
admin.site.register(Restaurant)
admin.site.register(Person)
admin.site.register(Child)



from celery import shared_task
import report_builder_scheduled.models


@shared_task
def report_builder_run_scheduled_report(scheduled_report_id: int):
    report = report_builder_scheduled.models.ScheduledReport.objects.get(pk=scheduled_report_id)
    report.run_report()

@shared_task
def report_builder_check_if_scheduled_report():
    """Run any reports that need run - this will kick off another task for
    the actual reports - so this should always run pretty fast """
    for scheduled in report_builder_scheduled.models.ScheduledReport.objects.filter(is_active=True):
        scheduled.run_from_schedule()


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import djmoney.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('demo_models', '0008_merge'),
    ]

    operations = [
        migrations.CreateModel(
            name='Comment',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, primary_key=True, verbose_name='ID')),
                ('object_pk', models.TextField()),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
        ),
        migrations.AlterField(
            model_name='account',
            name='balance',
            field=djmoney.models.fields.MoneyField(blank=True, default=None, default_currency='USD', null=True, decimal_places=2, max_digits=10),
        ),
        migrations.AlterField(
            model_name='account',
            name='balance_currency',
            field=djmoney.models.fields.CurrencyField(default='USD', editable=False, max_length=3, choices=[('AFN', 'Afghani'), ('DZD', 'Algerian Dinar'), ('ARS', 'Argentine Peso'), ('AMD', 'Armenian Dram'), ('AWG', 'Aruban Guilder'), ('AUD', 'Australian Dollar'), ('AZN', 'Azerbaijanian Manat'), ('BSD', 'Bahamian Dollar'), ('BHD', 'Bahraini Dinar'), ('THB', 'Baht'), ('BBD', 'Barbados Dollar'), ('BYR', 'Belarussian Ruble'), ('BZD', 'Belize Dollar'), ('BMD', 'Bermudian Dollar (customarily known as Bermuda Dollar)'), ('BTN', 'Bhutanese ngultrum'), ('VEF', 'Bolivar Fuerte'), ('XBA', 'Bond Markets Units European Composite Unit (EURCO)'), ('BRL', 'Brazilian Real'), ('BND', 'Brunei Dollar'), ('BGN', 'Bulgarian Lev'), ('BIF', 'Burundi Franc'), ('XOF', 'CFA Franc BCEAO'), ('XAF', 'CFA franc BEAC'), ('XPF', 'CFP Franc'), ('CAD', 'Canadian Dollar'), ('CVE', 'Cape Verde Escudo'), ('KYD', 'Cayman Islands Dollar'), ('CLP', 'Chilean peso'), ('XTS', 'Codes specifically reserved for testing purposes'), ('COP', 'Colombian peso'), ('KMF', 'Comoro Franc'), ('CDF', 'Congolese franc'), ('BAM', 'Convertible Marks'), ('NIO', 'Cordoba Oro'), ('CRC', 'Costa Rican Colon'), ('HRK', 'Croatian Kuna'), ('CUP', 'Cuban Peso'), ('CUC', 'Cuban convertible peso'), ('CZK', 'Czech Koruna'), ('GMD', 'Dalasi'), ('DKK', 'Danish Krone'), ('MKD', 'Denar'), ('DJF', 'Djibouti Franc'), ('STD', 'Dobra'), ('DOP', 'Dominican Peso'), ('VND', 'Dong'), ('XCD', 'East Caribbean Dollar'), ('EGP', 'Egyptian Pound'), ('ETB', 'Ethiopian Birr'), ('EUR', 'Euro'), ('XBB', 'European Monetary Unit (E.M.U.-6)'), ('XBD', 'European Unit of Account 17(E.U.A.-17)'), ('XBC', 'European Unit of Account 9(E.U.A.-9)'), ('FKP', 'Falkland Islands Pound'), ('FJD', 'Fiji Dollar'), ('HUF', 'Forint'), ('GHS', 'Ghana Cedi'), ('GIP', 'Gibraltar Pound'), ('XAU', 'Gold'), ('XFO', 'Gold-Franc'), ('PYG', 'Guarani'), ('GNF', 'Guinea Franc'), ('GYD', 'Guyana Dollar'), ('HTG', 'Haitian gourde'), ('HKD', 'Hong Kong Dollar'), ('UAH', 'Hryvnia'), ('ISK', 'Iceland Krona'), ('INR', 'Indian Rupee'), ('IRR', 'Iranian Rial'), ('IQD', 'Iraqi Dinar'), ('IMP', 'Isle of Man pount'), ('JMD', 'Jamaican Dollar'), ('JOD', 'Jordanian Dinar'), ('KES', 'Kenyan Shilling'), ('PGK', 'Kina'), ('LAK', 'Kip'), ('KWD', 'Kuwaiti Dinar'), ('AOA', 'Kwanza'), ('MMK', 'Kyat'), ('GEL', 'Lari'), ('LVL', 'Latvian Lats'), ('LBP', 'Lebanese Pound'), ('ALL', 'Lek'), ('HNL', 'Lempira'), ('SLL', 'Leone'), ('LSL', 'Lesotho loti'), ('LRD', 'Liberian Dollar'), ('LYD', 'Libyan Dinar'), ('SZL', 'Lilangeni'), ('LTL', 'Lithuanian Litas'), ('MGA', 'Malagasy Ariary'), ('MWK', 'Malawian Kwacha'), ('MYR', 'Malaysian Ringgit'), ('TMM', 'Manat'), ('MUR', 'Mauritius Rupee'), ('MZN', 'Metical'), ('MXN', 'Mexican peso'), ('MDL', 'Moldovan Leu'), ('MAD', 'Moroccan Dirham'), ('NGN', 'Naira'), ('ERN', 'Nakfa'), ('NAD', 'Namibian Dollar'), ('NPR', 'Nepalese Rupee'), ('ANG', 'Netherlands Antillian Guilder'), ('ILS', 'New Israeli Sheqel'), ('RON', 'New Leu'), ('TWD', 'New Taiwan Dollar'), ('NZD', 'New Zealand Dollar'), ('KPW', 'North Korean Won'), ('NOK', 'Norwegian Krone'), ('PEN', 'Nuevo Sol'), ('MRO', 'Ouguiya'), ('TOP', 'Paanga'), ('PKR', 'Pakistan Rupee'), ('XPD', 'Palladium'), ('MOP', 'Pataca'), ('PHP', 'Philippine Peso'), ('XPT', 'Platinum'), ('GBP', 'Pound Sterling'), ('BWP', 'Pula'), ('QAR', 'Qatari Rial'), ('GTQ', 'Quetzal'), ('ZAR', 'Rand'), ('OMR', 'Rial Omani'), ('KHR', 'Riel'), ('MVR', 'Rufiyaa'), ('IDR', 'Rupiah'), ('RUB', 'Russian Ruble'), ('RWF', 'Rwanda Franc'), ('XDR', 'SDR'), ('SHP', 'Saint Helena Pound'), ('SAR', 'Saudi Riyal'), ('RSD', 'Serbian Dinar'), ('SCR', 'Seychelles Rupee'), ('XAG', 'Silver'), ('SGD', 'Singapore Dollar'), ('SBD', 'Solomon Islands Dollar'), ('KGS', 'Som'), ('SOS', 'Somali Shilling'), ('TJS', 'Somoni'), ('LKR', 'Sri Lanka Rupee'), ('SDG', 'Sudanese Pound'), ('SRD', 'Surinam Dollar'), ('SEK', 'Swedish Krona'), ('CHF', 'Swiss Franc'), ('SYP', 'Syrian Pound'), ('BDT', 'Taka'), ('WST', 'Tala'), ('TZS', 'Tanzanian Shilling'), ('KZT', 'Tenge'), ('TTD', 'Trinidad and Tobago Dollar'), ('MNT', 'Tugrik'), ('TND', 'Tunisian Dinar'), ('TRY', 'Turkish Lira'), ('TVD', 'Tuvalu dollar'), ('AED', 'UAE Dirham'), ('XFU', 'UIC-Franc'), ('USD', 'US Dollar'), ('UGX', 'Uganda Shilling'), ('UYU', 'Uruguayan peso'), ('UZS', 'Uzbekistan Sum'), ('VUV', 'Vatu'), ('KRW', 'Won'), ('YER', 'Yemeni Rial'), ('JPY', 'Yen'), ('CNY', 'Yuan Renminbi'), ('ZMK', 'Zambian Kwacha'), ('ZMW', 'Zambian Kwacha'), ('ZWD', 'Zimbabwe Dollar A/06'), ('ZWN', 'Zimbabwe dollar A/08'), ('ZWL', 'Zimbabwe dollar A/09'), ('PLN', 'Zloty')]),
        ),
        migrations.AlterField(
            model_name='account',
            name='budget',
            field=djmoney.models.fields.MoneyField(blank=True, default=None, default_currency='USD', null=True, decimal_places=4, max_digits=20),
        ),
        migrations.AlterField(
            model_name='account',
            name='budget_currency',
            field=djmoney.models.fields.CurrencyField(default='USD', editable=False, max_length=3, choices=[('AFN', 'Afghani'), ('DZD', 'Algerian Dinar'), ('ARS', 'Argentine Peso'), ('AMD', 'Armenian Dram'), ('AWG', 'Aruban Guilder'), ('AUD', 'Australian Dollar'), ('AZN', 'Azerbaijanian Manat'), ('BSD', 'Bahamian Dollar'), ('BHD', 'Bahraini Dinar'), ('THB', 'Baht'), ('BBD', 'Barbados Dollar'), ('BYR', 'Belarussian Ruble'), ('BZD', 'Belize Dollar'), ('BMD', 'Bermudian Dollar (customarily known as Bermuda Dollar)'), ('BTN', 'Bhutanese ngultrum'), ('VEF', 'Bolivar Fuerte'), ('XBA', 'Bond Markets Units European Composite Unit (EURCO)'), ('BRL', 'Brazilian Real'), ('BND', 'Brunei Dollar'), ('BGN', 'Bulgarian Lev'), ('BIF', 'Burundi Franc'), ('XOF', 'CFA Franc BCEAO'), ('XAF', 'CFA franc BEAC'), ('XPF', 'CFP Franc'), ('CAD', 'Canadian Dollar'), ('CVE', 'Cape Verde Escudo'), ('KYD', 'Cayman Islands Dollar'), ('CLP', 'Chilean peso'), ('XTS', 'Codes specifically reserved for testing purposes'), ('COP', 'Colombian peso'), ('KMF', 'Comoro Franc'), ('CDF', 'Congolese franc'), ('BAM', 'Convertible Marks'), ('NIO', 'Cordoba Oro'), ('CRC', 'Costa Rican Colon'), ('HRK', 'Croatian Kuna'), ('CUP', 'Cuban Peso'), ('CUC', 'Cuban convertible peso'), ('CZK', 'Czech Koruna'), ('GMD', 'Dalasi'), ('DKK', 'Danish Krone'), ('MKD', 'Denar'), ('DJF', 'Djibouti Franc'), ('STD', 'Dobra'), ('DOP', 'Dominican Peso'), ('VND', 'Dong'), ('XCD', 'East Caribbean Dollar'), ('EGP', 'Egyptian Pound'), ('ETB', 'Ethiopian Birr'), ('EUR', 'Euro'), ('XBB', 'European Monetary Unit (E.M.U.-6)'), ('XBD', 'European Unit of Account 17(E.U.A.-17)'), ('XBC', 'European Unit of Account 9(E.U.A.-9)'), ('FKP', 'Falkland Islands Pound'), ('FJD', 'Fiji Dollar'), ('HUF', 'Forint'), ('GHS', 'Ghana Cedi'), ('GIP', 'Gibraltar Pound'), ('XAU', 'Gold'), ('XFO', 'Gold-Franc'), ('PYG', 'Guarani'), ('GNF', 'Guinea Franc'), ('GYD', 'Guyana Dollar'), ('HTG', 'Haitian gourde'), ('HKD', 'Hong Kong Dollar'), ('UAH', 'Hryvnia'), ('ISK', 'Iceland Krona'), ('INR', 'Indian Rupee'), ('IRR', 'Iranian Rial'), ('IQD', 'Iraqi Dinar'), ('IMP', 'Isle of Man pount'), ('JMD', 'Jamaican Dollar'), ('JOD', 'Jordanian Dinar'), ('KES', 'Kenyan Shilling'), ('PGK', 'Kina'), ('LAK', 'Kip'), ('KWD', 'Kuwaiti Dinar'), ('AOA', 'Kwanza'), ('MMK', 'Kyat'), ('GEL', 'Lari'), ('LVL', 'Latvian Lats'), ('LBP', 'Lebanese Pound'), ('ALL', 'Lek'), ('HNL', 'Lempira'), ('SLL', 'Leone'), ('LSL', 'Lesotho loti'), ('LRD', 'Liberian Dollar'), ('LYD', 'Libyan Dinar'), ('SZL', 'Lilangeni'), ('LTL', 'Lithuanian Litas'), ('MGA', 'Malagasy Ariary'), ('MWK', 'Malawian Kwacha'), ('MYR', 'Malaysian Ringgit'), ('TMM', 'Manat'), ('MUR', 'Mauritius Rupee'), ('MZN', 'Metical'), ('MXN', 'Mexican peso'), ('MDL', 'Moldovan Leu'), ('MAD', 'Moroccan Dirham'), ('NGN', 'Naira'), ('ERN', 'Nakfa'), ('NAD', 'Namibian Dollar'), ('NPR', 'Nepalese Rupee'), ('ANG', 'Netherlands Antillian Guilder'), ('ILS', 'New Israeli Sheqel'), ('RON', 'New Leu'), ('TWD', 'New Taiwan Dollar'), ('NZD', 'New Zealand Dollar'), ('KPW', 'North Korean Won'), ('NOK', 'Norwegian Krone'), ('PEN', 'Nuevo Sol'), ('MRO', 'Ouguiya'), ('TOP', 'Paanga'), ('PKR', 'Pakistan Rupee'), ('XPD', 'Palladium'), ('MOP', 'Pataca'), ('PHP', 'Philippine Peso'), ('XPT', 'Platinum'), ('GBP', 'Pound Sterling'), ('BWP', 'Pula'), ('QAR', 'Qatari Rial'), ('GTQ', 'Quetzal'), ('ZAR', 'Rand'), ('OMR', 'Rial Omani'), ('KHR', 'Riel'), ('MVR', 'Rufiyaa'), ('IDR', 'Rupiah'), ('RUB', 'Russian Ruble'), ('RWF', 'Rwanda Franc'), ('XDR', 'SDR'), ('SHP', 'Saint Helena Pound'), ('SAR', 'Saudi Riyal'), ('RSD', 'Serbian Dinar'), ('SCR', 'Seychelles Rupee'), ('XAG', 'Silver'), ('SGD', 'Singapore Dollar'), ('SBD', 'Solomon Islands Dollar'), ('KGS', 'Som'), ('SOS', 'Somali Shilling'), ('TJS', 'Somoni'), ('LKR', 'Sri Lanka Rupee'), ('SDG', 'Sudanese Pound'), ('SRD', 'Surinam Dollar'), ('SEK', 'Swedish Krona'), ('CHF', 'Swiss Franc'), ('SYP', 'Syrian Pound'), ('BDT', 'Taka'), ('WST', 'Tala'), ('TZS', 'Tanzanian Shilling'), ('KZT', 'Tenge'), ('TTD', 'Trinidad and Tobago Dollar'), ('MNT', 'Tugrik'), ('TND', 'Tunisian Dinar'), ('TRY', 'Turkish Lira'), ('TVD', 'Tuvalu dollar'), ('AED', 'UAE Dirham'), ('XFU', 'UIC-Franc'), ('USD', 'US Dollar'), ('UGX', 'Uganda Shilling'), ('UYU', 'Uruguayan peso'), ('UZS', 'Uzbekistan Sum'), ('VUV', 'Vatu'), ('KRW', 'Won'), ('YER', 'Yemeni Rial'), ('JPY', 'Yen'), ('CNY', 'Yuan Renminbi'), ('ZMK', 'Zambian Kwacha'), ('ZMW', 'Zambian Kwacha'), ('ZWD', 'Zimbabwe Dollar A/06'), ('ZWN', 'Zimbabwe dollar A/08'), ('ZWL', 'Zimbabwe dollar A/09'), ('PLN', 'Zloty')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='DisplayField',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('path', models.CharField(max_length=2000, blank=True)),
                ('path_verbose', models.CharField(max_length=2000, blank=True)),
                ('field', models.CharField(max_length=2000)),
                ('field_verbose', models.CharField(max_length=2000)),
                ('name', models.CharField(max_length=2000)),
                ('sort', models.IntegerField(null=True, blank=True)),
                ('sort_reverse', models.BooleanField(default=False, verbose_name=b'Reverse')),
                ('width', models.IntegerField(default=15)),
                ('aggregate', models.CharField(blank=True, max_length=5, choices=[(b'Sum', b'Sum'), (b'Count', b'Count'), (b'Avg', b'Avg'), (b'Max', b'Max'), (b'Min', b'Min')])),
                ('position', models.PositiveSmallIntegerField(null=True, blank=True)),
                ('total', models.BooleanField(default=False)),
                ('group', models.BooleanField(default=False)),
            ],
            options={
                'ordering': ['position'],
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='FilterField',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('path', models.CharField(max_length=2000, blank=True)),
                ('path_verbose', models.CharField(max_length=2000, blank=True)),
                ('field', models.CharField(max_length=2000)),
                ('field_verbose', models.CharField(max_length=2000)),
                ('filter_type', models.CharField(default=b'icontains', max_length=20, blank=True, choices=[(b'exact', b'Equals'), (b'iexact', b'Equals (case-insensitive)'), (b'contains', b'Contains'), (b'icontains', b'Contains (case-insensitive)'), (b'in', b'in (comma seperated 1,2,3)'), (b'gt', b'Greater than'), (b'gte', b'Greater than equals'), (b'lt', b'Less than'), (b'lte', b'Less than equals'), (b'startswith', b'Starts with'), (b'istartswith', b'Starts with (case-insensitive)'), (b'endswith', b'Ends with'), (b'iendswith', b'Ends with  (case-insensitive)'), (b'range', b'range'), (b'week_day', b'Week day'), (b'isnull', b'Is null'), (b'regex', b'Regular Expression'), (b'iregex', b'Reg. Exp. (case-insensitive)')])),
                ('filter_value', models.CharField(max_length=2000)),
                ('filter_value2', models.CharField(max_length=2000, blank=True)),
                ('exclude', models.BooleanField(default=False)),
                ('position', models.PositiveSmallIntegerField(null=True, blank=True)),
            ],
            options={
                'ordering': ['position'],
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Format',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(default=b'', max_length=50, blank=True)),
                ('string', models.CharField(default=b'', help_text=b'Python string format. Ex ${} would place a $ in front of the result.', max_length=300, blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Report',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('slug', models.SlugField(verbose_name=b'Short Name')),
                ('description', models.TextField(blank=True)),
                ('created', models.DateField(auto_now_add=True)),
                ('modified', models.DateField(auto_now=True)),
                ('distinct', models.BooleanField(default=False)),
                ('report_file', models.FileField(upload_to=b'report_files', blank=True)),
                ('report_file_creation', models.DateTimeField(null=True, blank=True)),
                ('root_model', models.ForeignKey(to='contenttypes.ContentType')),
                ('starred', models.ManyToManyField(help_text=b'These users have starred this report for easy reference.', related_name='report_starred_set', to=settings.AUTH_USER_MODEL, blank=True)),
                ('user_created', models.ForeignKey(blank=True, editable=False, to=settings.AUTH_USER_MODEL, null=True)),
                ('user_modified', models.ForeignKey(related_name='report_modified_set', blank=True, editable=False, to=settings.AUTH_USER_MODEL, null=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='filterfield',
            name='report',
            field=models.ForeignKey(to='report_builder.Report'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='displayfield',
            name='display_format',
            field=models.ForeignKey(blank=True, to='report_builder.Format', null=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='displayfield',
            name='report',
            field=models.ForeignKey(to='report_builder.Report'),
            preserve_default=True,
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.conf import settings
from django.test import TestCase

from ..utils import (
    get_properties_from_model, get_direct_fields_from_model,
    get_relation_fields_from_model, get_model_from_path_string)
from ..mixins import GetFieldsMixin
from ..models import Report, DisplayField, FilterField
from report_builder_demo.demo_models.models import (
    Bar, Restaurant, Waiter, Comment, Place)


class RelationUtilityFunctionTests(TestCase):
    def test_a_initial_rel_field_name(self):
        """
        Test that the initial assumption about the ManyToOneRel field_name is
        correct
        """
        self.assertEquals(Waiter.restaurant.field.rel.field_name, "place")

    def test_get_relation_fields_from_model_does_not_change_field_name(self):
        """
        Make sure that getting related_fields doesn't overwrite field_name

        Waiter has a ForeignKey to Restaurant.
        The relation from Restaurant to Waiter is a ManyToOneRel object.
        'place' is the PK of Restaurant. The ManyToOneRel field_name should be
        the same at the PK, unless to_field is set on the ForeignKey.

        ManyToManyRel objects are not affected.
        """
        get_relation_fields_from_model(Restaurant)
        self.assertEquals(Waiter.restaurant.field.rel.field_name, "place")
        Waiter.restaurant.field.rel.get_related_field()


class UtilityFunctionTests(TestCase):

    def setUp(self):
        self.report_ct = ContentType.objects.get_for_model(Report)
        self.report = Report.objects.create(
            name="foo report",
            root_model=self.report_ct)
        self.filter_field = FilterField.objects.create(
            report=self.report,
            field="X",
            field_verbose="stuff",
            filter_type='contains',
            filter_value='Lots of spam')

    def get_fields_names(self, fields):
        return [field.name for field in fields]

    def test_get_relation_fields_from_model(self):
        fields = get_relation_fields_from_model(Report)
        names = self.get_fields_names(fields)
        self.assertTrue('displayfield' in names or 'report_builder:displayfield' in names)
        self.assertTrue('filterfield' in names or 'report_builder:filterfield' in names)
        self.assertTrue('root_model' in names)
        self.assertEquals(len(names), 7)

    def test_get_model_from_path_string(self):
        result = get_model_from_path_string(Restaurant, 'waiter__name')
        self.assertEqual(result, Waiter)

    def test_get_model_from_path_string_one_to_one(self):
        """Test that one-to-one relationships don't break this function"""
        result = get_model_from_path_string(Restaurant, 'place__serves_pizza')
        self.assertEqual(result, Place)


    def test_get_direct_fields_from_model(self):
        fields = get_direct_fields_from_model(Report)
        names = self.get_fields_names(fields)
        self.assertTrue('created' in names)
        self.assertTrue('description' in names)
        self.assertTrue('distinct' in names)
        self.assertTrue('id' in names)
        self.assertEquals(len(names), 9)

    def test_get_fields(self):
        """ Test GetFieldsMixin.get_fields """
        obj = GetFieldsMixin()
        obj.get_fields(
            Bar,
            "foos",
        )

    def test_get_gfk_fields_from_model(self):
        fields = get_direct_fields_from_model(Comment)

    def test_get_properties_from_model(self):
        properties = get_properties_from_model(DisplayField)
        self.assertEquals(properties[0]['label'], 'choices')
        self.assertEquals(properties[1]['label'], 'choices_dict')

    def test_filter_property(self):
        # Not a very complete test - only tests one type of filter
        result = self.filter_field.filter_property('spam')
        self.assertTrue(result)

    def test_custom_global_model_manager(self):
        """ test for custom global model manager """
        if getattr(settings, 'REPORT_BUILDER_MODEL_MANAGER', False):
            self.assertEquals(
                self.report._get_model_manager(),
                settings.REPORT_BUILDER_MODEL_MANAGER)

    def test_custom_model_manager(self):
        """ test for custom model manager """
        if getattr(
                self.report.root_model.model_class(),
                'report_builder_model_manager',
                True
        ):
            # change setup to use actual field and value
            self.filter_field.field = 'name'
            self.filter_field.filter_value = 'foo'
            self.filter_field.save()
            # coverage of get_query
            objects = self.report.get_query()
            # expect custom manager to return correct object with filters
            self.assertEquals(objects[0], self.report)



from django.contrib.auth import get_user_model
from django.conf import settings
from django.core import mail
from django.test import TestCase
from django.test.utils import override_settings

from model_mommy import mommy
from report_builder.tasks import report_builder_file_async_report_save


from ..email import email_report


User = get_user_model()


class ViewTests(TestCase):
    def test_email_report_without_template(self):
        settings.REPORT_BUILDER_EMAIL_NOTIFICATION = True
        email_subject = getattr(settings, 'REPORT_BUILDER_EMAIL_SUBJECT', False) or "Report is ready"
        user = User.objects.get_or_create(username='example', email='to@example.com')[0]
        email_report(email_subject, user)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, email_subject)
        settings.REPORT_BUILDER_EMAIL_NOTIFICATION = None
        mail.outbox = []

    def test_email_report_with_template(self):
        settings.REPORT_BUILDER_EMAIL_NOTIFICATION = True
        report_url = 'http://fakeurl.com/fakestuffs'
        username = 'example'
        email_subject = getattr(settings, 'REPORT_BUILDER_EMAIL_SUBJECT', False) or "Report is ready"
        user = User.objects.get_or_create(username=username, email='to@example.com')[0]
        email_report(report_url, user)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, email_subject)
        self.assertEqual(mail.outbox[0].alternatives[0][0], "<p>Hello {0},</p>\n<br>\n<p>The report is <a href='{1}'>here</u></p>".format(username, report_url))
        settings.REPORT_BUILDER_EMAIL_NOTIFICATION = None
        settings.REPORT_BUILDER_EMAIL_TEMPLATE = None
        mail.outbox = []

    @override_settings(CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
    def test_report_builder_file_async_report_save(self):
        """
        Test to ensure the celery task will succeed given proper params in the right order.
        """
        user = User.objects.create(username='testy', is_staff=True, is_superuser=True)
        report = mommy.make('Report')
        res = report_builder_file_async_report_save.delay(report.id, user.id, 'xlsx')
        self.assertEqual(res.successful(), True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Bar',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, verbose_name='ID', primary_key=True)),
                ('char_field', models.CharField(blank=True, max_length=50)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Foo',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, verbose_name='ID', primary_key=True)),
                ('char_field', models.CharField(blank=True, max_length=50)),
                ('char_field2', models.CharField(blank=True, max_length=50)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='FooExclude',
            fields=[
                ('foo_ptr', models.OneToOneField(auto_created=True, serialize=False, parent_link=True, primary_key=True, to='demo_models.Foo')),
            ],
            options={
            },
            bases=('demo_models.foo',),
        ),
    ]



from celery import shared_task


@shared_task
def report_builder_file_async_report_save(report_id, user_id, file_type):
    """ Start a report task """
    from .views import DownloadFileView
    view = DownloadFileView()
    view.process_report(report_id, user_id, file_type, to_response=False)



from django.db import models
from django.contrib.contenttypes.fields import GenericForeignKey
from django.utils.functional import cached_property

from djmoney.models.fields import MoneyField


class Foo(models.Model):
    char_field = models.CharField(max_length=50, blank=True)
    char_field2 = models.CharField(max_length=50, blank=True)

    class ReportBuilder:
        fields = ('char_field',)


class FooExclude(Foo):
    class ReportBuilder:
        exclude = ('char_field2',)


class Bar(models.Model):
    char_field = models.CharField(max_length=50, blank=True)
    foos = models.ManyToManyField(Foo, blank=True)

    CHECK = 'CH'
    MATE = 'MA'
    CHESS_CHOICES = (
        (CHECK, 'CHECK'),
        (MATE, 'CHECKMATE'),
    )

    check_mate_status = models.CharField(
        max_length=2,
        choices=CHESS_CHOICES,
        default=CHECK
    )

    @property
    def i_want_char_field(self):
        return 'lol no'

    @cached_property
    def i_need_char_field(self):
        return 'lol yes'

    class ReportBuilder:
        extra = ('i_want_char_field', 'i_need_char_field',)
        filters = ('char_field',)


class Account(models.Model):
    name = models.CharField(max_length=255, blank=True, null=True, db_index=True)
    balance = MoneyField(max_digits=10, decimal_places=2, default_currency='USD', blank=True, null=True)
    budget = MoneyField(max_digits=20, decimal_places=4, default_currency='USD', blank=True, null=True)

    class ReportBuilder:
        fields = ('budget', 'id', 'name', 'balance',)
        filters = ('budget',)
        defaults = ('budget',)
        exclude = ('balance_currency',)


class Place(models.Model):
    name = models.CharField(max_length=50)
    address = models.CharField(max_length=80)

    def __str__(self):
        return "%s the place" % self.name


class Restaurant(models.Model):
    place = models.OneToOneField(Place, primary_key=True)
    serves_hot_dogs = models.BooleanField(default=False)
    serves_pizza = models.BooleanField(default=False)

    def __str__(self):
        return "%s the restaurant" % self.place.name


class Waiter(models.Model):
    restaurant = models.ForeignKey(Restaurant)
    name = models.CharField(max_length=50)
    days_worked = models.IntegerField(blank=True, null=True, default=None)

    def __str__(self):
        return "%s the waiter at %s" % (self.name, self.restaurant)


class Person(models.Model):
    first_name = models.CharField(max_length=50)
    last_name = models.CharField(max_length=50)
    last_login = models.DateField(
        auto_now=True)
    last_modifed = models.DateField(
        null=True,
        blank=True)
    birth_date = models.DateTimeField(
        null=True,
        blank=True)
    hammer_time = models.TimeField(
        null=True,
        blank=True,
        help_text="U Can't Touch This")


class Child(models.Model):
    parent = models.ForeignKey(Person, related_name='children')
    first_name = models.CharField(max_length=50)
    last_name = models.CharField(max_length=50)
    age = models.IntegerField(null=True, blank=True, default=None)
    color = models.CharField(max_length=1, blank=True, default='', choices=(
        ('R', 'Red'),
        ('G', 'Green'),
        ('B', 'Blue'),
    ))


class Comment(models.Model):
    """ django-contrib-comments like model """
    content_type = models.ForeignKey('contenttypes.ContentType')
    object_pk = models.TextField()
    content_object = GenericForeignKey(
        ct_field="content_type", fk_field="object_pk")



import copy
from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404
from django.http import JsonResponse
from django.utils.functional import cached_property
from django.conf import settings
from rest_framework import viewsets, status
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.decorators import detail_route
from rest_framework.response import Response
from rest_framework.permissions import IsAdminUser
from ..models import Report, Format, FilterField
from .serializers import (
    ReportNestedSerializer, ReportSerializer, FormatSerializer,
    FilterFieldSerializer, ContentTypeSerializer)
from ..mixins import GetFieldsMixin, DataExportMixin
from django.core import serializers
from ..utils import duplicate


def find_exact_position(fields_list, item):
    current_position = 0
    for i in fields_list:
        if (i.name == item.name and
                i.get_internal_type() == item.get_internal_type()):
            return current_position
        current_position += 1
    return -1


class ReportBuilderViewMixin:
    """ Set up explicit settings so that project defaults
    don't interfer with report builder's api. """
    permission_classes = (IsAdminUser,)
    pagination_class = None

class ConfigView(ReportBuilderViewMixin, APIView):
    def get(self, request):
        data = {
            'async_report': getattr( settings, 'REPORT_BUILDER_ASYNC_REPORT', False ),
            'formats': FormatSerializer(Format.objects.all(), many=True).data
        }
        return JsonResponse(data)


class FormatViewSet(ReportBuilderViewMixin, viewsets.ModelViewSet):
    queryset = Format.objects.all()
    serializer_class = FormatSerializer


class FilterFieldViewSet(ReportBuilderViewMixin, viewsets.ModelViewSet):
    queryset = FilterField.objects.all()
    serializer_class = FilterFieldSerializer


class ContentTypeViewSet(ReportBuilderViewMixin, viewsets.ReadOnlyModelViewSet):
    """ Read only view of content types.
    Used to populate choices for new report root model.
    """
    queryset = ContentType.objects.all()
    serializer_class = ContentTypeSerializer


class ReportViewSet(ReportBuilderViewMixin, viewsets.ModelViewSet):
    queryset = Report.objects.all()
    serializer_class = ReportSerializer


class ReportNestedViewSet(ReportBuilderViewMixin, viewsets.ModelViewSet):
    queryset = Report.objects.all()
    serializer_class = ReportNestedSerializer

    def perform_create(self, serializer):
        serializer.save(user_created=self.request.user)

    def perform_update(self, serializer):
        serializer.save(user_modified=self.request.user)

    @detail_route(methods=['post'])
    def copy_report(self, request, pk=None):
        report = self.get_object()
        new_report = duplicate(report, changes=(
            ('name', '{0} (copy)'.format(report.name)),
            ('user_created', request.user),
            ('user_modified', request.user),
        ))

        # duplicate does not get related
        for display in report.displayfield_set.all():
            new_display = copy.copy(display)
            new_display.pk = None
            new_display.report = new_report
            new_display.save()
        for report_filter in report.filterfield_set.all():
            new_filter = copy.copy(report_filter)
            new_filter.pk = None
            new_filter.report = new_report
            new_filter.save()

        serializer = ReportNestedSerializer(new_report)
        return JsonResponse(serializer.data)

        


class RelatedFieldsView(ReportBuilderViewMixin, GetFieldsMixin, APIView):
    """ Get related fields from an ORM model """
    def get_data_from_request(self, request):
        self.model = request.data['model']
        self.path = request.data['path']
        self.path_verbose = request.data.get('path_verbose', '')
        self.field = request.data['field']
        self.model_class = ContentType.objects.get(pk=self.model).model_class()

    def post(self, request):
        self.get_data_from_request(request)
        new_fields, model_ct, path = self.get_related_fields(
            self.model_class,
            self.field,
            self.path,
            self.path_verbose,)
        result = []
        for new_field in new_fields:
            included_model = True
            split_name = new_field.name.split(':')
            if len(split_name) == 1:
                split_name.append('')
                split_name[1] = split_name[0]
                split_name[0] = False
                model_information = split_name[1]
            else:
                model_information = split_name[0] + "." + split_name[1]
            app_label = split_name[0]
            model_name = split_name[1]
            if getattr(settings, 'REPORT_BUILDER_INCLUDE', False):
                includes = getattr(settings, 'REPORT_BUILDER_INCLUDE')
                # If it is not included as 'foo' and not as 'demo_models.foo'
                if (model_name not in includes and
                        model_information not in includes):
                    included_model = False
            if getattr(settings, 'REPORT_BUILDER_EXCLUDE', False):
                excludes = getattr(settings, 'REPORT_BUILDER_EXCLUDE')
                # If it is excluded as 'foo' and as 'demo_models.foo'
                if (model_name in excludes or model_information in excludes):
                    included_model = False
            verbose_name = getattr(new_field, 'verbose_name', None)
            if verbose_name is None:
                verbose_name = new_field.get_accessor_name()
            result += [{
                'field_name': new_field.field_name,
                'verbose_name': verbose_name,
                'path': path,
                'help_text': getattr(new_field, 'help_text', ''),
                'model_id': model_ct.id,
                'parent_model_name': model_name,
                'parent_model_app_label': app_label,
                'included_model': included_model,
            }]
        return Response(result)


class FieldsView(RelatedFieldsView):
    """ Get direct fields and properties on an ORM model
    """
    def post(self, request):
        self.get_data_from_request(request)
        field_data = self.get_fields(
            self.model_class,
            self.field,
            self.path,
            self.path_verbose,)

        # External packages might cause duplicates. This clears it up
        new_set = []
        for i in field_data['fields']:
            if i not in new_set:
                new_set.append(i)
        field_data['fields'] = new_set

        result = []
        fields = None
        filters = None
        extra = None
        defaults = None
        meta = getattr(self.model_class, 'ReportBuilder', None)
        if meta is not None:
            fields = getattr(meta, 'fields', None)
            filters = getattr(meta, 'filters', None)
            exclude = getattr(meta, 'exclude', None)
            extra = getattr(meta, 'extra', None)
            defaults = getattr(meta, 'defaults', None)
            if fields is not None:
                fields = list(fields)
                for field in copy.copy(field_data['fields']):
                    if field.name not in fields:
                        index = find_exact_position(
                            field_data['fields'],
                            field
                        )
                        if index != -1:
                            field_data['fields'].pop(index)
            if exclude is not None:
                for field in copy.copy(field_data['fields']):
                    if field.name in exclude:
                        index = find_exact_position(
                            field_data['fields'],
                            field
                        )
                        if index != -1:
                            field_data['fields'].pop(index)
            if extra is not None:
                extra = list(extra)

        for new_field in field_data['fields']:
            verbose_name = getattr(new_field, 'verbose_name', None)
            if not verbose_name:
                verbose_name = new_field.get_accessor_name()
            result += [{
                'name': new_field.name,
                'field': new_field.name,
                'field_verbose': verbose_name,
                'field_type': new_field.get_internal_type(),
                'is_default': True if defaults is None or
                new_field.name in defaults else False,
                'field_choices': new_field.choices,
                'can_filter': True if filters is None or
                new_field.name in filters else False,
                'path': field_data['path'],
                'path_verbose': field_data['path_verbose'],
                'help_text': new_field.help_text,
            }]
        # Add properties
        if fields is not None or extra is not None:
            if fields and extra:
                extra_fields = list(set(fields + extra))
            elif fields is not None:
                extra_fields = fields
            else:
                extra_fields = extra
            for field in extra_fields:
                field_attr = getattr(self.model_class, field, None)
                if isinstance(field_attr, (property, cached_property)):
                    result += [{
                        'name': field,
                        'field': field,
                        'field_verbose': field,
                        'field_type': 'Property',
                        'field_choices': None,
                        'can_filter': True if filters is None or
                        field in filters else False,
                        'path': field_data['path'],
                        'path_verbose': field_data['path_verbose'],
                        'is_default': True if defaults is None or
                        field in defaults else False,
                        'help_text': 'Adding this property will '
                        'significantly increase the time it takes to run a '
                        'report.'
                    }]
        # Add custom fields
        custom_fields = field_data.get('custom_fields', None)
        if custom_fields:
            for field in custom_fields:
                result += [{
                    'name': field.name,
                    'field': field.name,
                    'field_verbose': field.name,
                    'field_type': 'Custom Field',
                    'field_choices': getattr(field, 'choices', None),
                    'can_filter': True if filters is None or
                    field.name in filters else False,
                    'path': field_data['path'],
                    'path_verbose': field_data['path_verbose'],
                    'is_default': True if defaults is None or
                    field.name in defaults else False,
                    'help_text': 'This is a custom field.',
                }]
        return Response(result)


class GenerateReport(ReportBuilderViewMixin, DataExportMixin, APIView):
    def get(self, request, report_id=None):
        return self.post(request, report_id=report_id)

    def post(self, request, report_id=None):
        report = get_object_or_404(Report, pk=report_id)

        objects_list = report.report_to_list(
            user=request.user,
            preview=True,)
        display_fields = report.get_good_display_fields().values_list(
            'name', flat=True)
        response = {
            'data': objects_list,
            'meta': {'titles': display_fields},
        }

        return Response(response)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-09-15 20:46
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('report_builder', '0004_auto_20170915_2046'),
        ('django_celery_beat', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ScheduledReport',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True)),
                ('other_emails', models.CharField(blank=True, help_text='comma separated list of emails to send to in addition to users', max_length=1000)),
                ('last_run_at', models.DateTimeField(blank=True, editable=False, null=True)),
                ('crontab', models.ForeignKey(blank=True, help_text='Use one of interval/crontab', null=True, on_delete=django.db.models.deletion.CASCADE, to='django_celery_beat.CrontabSchedule', verbose_name='crontab')),
                ('interval', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='django_celery_beat.IntervalSchedule', verbose_name='interval')),
                ('report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='report_builder.Report')),
                ('users', models.ManyToManyField(blank=True, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from django.contrib import admin
from django.core.urlresolvers import reverse
from .models import ScheduledReport


@admin.register(ScheduledReport)
class ScheduledReportAdmin(admin.ModelAdmin):
    list_display = ('report', 'is_active', 'last_run_at', 'run_report_url')
    list_filter = ('is_active', 'last_run_at')
    readonly_fields = ('last_run_at',)

    def run_report_url(self, obj):
        url = reverse('run_scheduled_report', kwargs={'pk': obj.id})
        return '<a href="%s">Run</a>' % (url,)

    run_report_url.allow_tags = True
    run_report_url.short_description = ''



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-11-21 03:46
from __future__ import unicode_literals

from django.db import migrations, models
import djmoney.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('demo_models', '0009_auto_20151209_2136'),
    ]

    operations = [
        migrations.AddField(
            model_name='person',
            name='birth_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='person',
            name='hammer_time',
            field=models.TimeField(blank=True, help_text="U Can't Touch This", null=True),
        ),
        migrations.AddField(
            model_name='person',
            name='last_login',
            field=models.DateField(auto_now=True),
        ),
        migrations.AddField(
            model_name='person',
            name='last_modifed',
            field=models.DateField(blank=True, null=True),
        ),
    ]



from setuptools import setup, find_packages

setup(
    name="django-report-builder",
    version="5.0.0-beta7",
    author="David Burke",
    author_email="david@burkesoftware.com",
    description=("Query and Report builder for Django ORM"),
    license="BSD",
    keywords="django report",
    url="https://gitlab.com/burke-software/django-report-builder",
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        'Environment :: Web Environment',
        'Framework :: Django',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        "License :: OSI Approved :: BSD License",
    ],
    python_requires='>=3.3',
    install_requires=[
        'openpyxl >= 2.2.1',
        'python-dateutil',
        'djangorestframework>=3.1.0',
        "six"
    ]
)



from django.conf.urls import include, url
from django.conf import settings
from django.conf.urls.static import static

from django.contrib import admin
admin.autodiscover()

urlpatterns = [
    url(r'^admin/', include(admin.site.urls)),
    url(r'^report_builder/', include('report_builder_scheduled.urls')),
    url(r'^report_builder/', include('report_builder.urls')),
]
if settings.DEBUG:
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
import report_builder_scheduled.tasks


AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


class ScheduledReport(models.Model):
    """ A scheduled report that runs and emails itself to various users on
    a recurring basis. Requires celery. """
    is_active = models.BooleanField(default=True)
    report = models.ForeignKey('report_builder.Report')
    users = models.ManyToManyField(
            AUTH_USER_MODEL,
            limit_choices_to={'is_staff': True},
            blank=True,
            help_text="Staff users to notify")
    other_emails = models.CharField(
        max_length=1000,
        blank=True,
        help_text="comma separated list of emails to send to in addition to users")
    last_run_at = models.DateTimeField(auto_now_add=True, editable=False)
    interval = models.ForeignKey(
        'django_celery_beat.IntervalSchedule', on_delete=models.CASCADE,
        null=True, blank=True, verbose_name=_('interval'),
    )
    crontab = models.ForeignKey(
        'django_celery_beat.CrontabSchedule', on_delete=models.CASCADE, null=True, blank=True,
        verbose_name=_('crontab'), help_text=_('Use one of interval/crontab'),
    )

    def __str__(self):
        return str(self.report)

    def _get_list_of_emails(self) -> [str]:
        """ Get list of emails for all users to return """
        emails = list(self.users.exclude(email="").values_list('email', flat=True))
        emails += [x.strip() for x in self.other_emails.split(',') if x != '']
        return emails

    def _is_due(self) -> bool:
        """Check if due to run, check against both cron and interval """
        if self.interval:
            is_due, next = self.interval.schedule.is_due(self.last_run_at)
            if is_due:
                return True
        if self.crontab:
            is_due, next = self.crontab.schedule.is_due(self.last_run_at)
            return is_due
        return False

    def run_report(self):
        self.report.run_report('xlsx', scheduled=True, email_to=self._get_list_of_emails())
        self.last_run_at = timezone.now()
        self.save()

    def run_from_schedule(self):
        """Run this only from a celery task - check if report needs to run and
        if so run it as it's own celery task """
        if self._is_due():
            report_builder_scheduled.tasks.report_builder_run_scheduled_report.delay(self.id)



from django.conf.urls import url, include
from django.contrib.admin.views.decorators import staff_member_required
from rest_framework import routers
from . import views
from .api import views as api_views
from django.conf import settings

router = routers.DefaultRouter()
router.register(r'reports', api_views.ReportViewSet)
router.register(r'report', api_views.ReportNestedViewSet)
router.register(r'formats', api_views.FormatViewSet)
router.register(r'filterfields', api_views.FilterFieldViewSet)
router.register(r'contenttypes', api_views.ContentTypeViewSet)

urlpatterns = [
    url(r'^report/(?P<pk>\d+)/download_file/$', views.DownloadFileView.as_view(), name="report_download_file"),
    url(r'^report/(?P<pk>\d+)/download_file/(?P<filetype>.+)/$', views.DownloadFileView.as_view(), name="report_download_file"),
    url(r'^report/(?P<pk>\d+)/check_status/(?P<task_id>.+)/$', views.check_status, name="report_check_status"),
    url(r'^report/(?P<pk>\d+)/add_star/$', views.ajax_add_star, name="ajax_add_star"),
    url(r'^report/(?P<pk>\d+)/create_copy/$', views.create_copy, name="report_builder_create_copy"),
    url(r'^export_to_report/$', views.ExportToReport.as_view(), name="export_to_report"),
    url(r'^api/', include(router.urls)),
    url(r'^api/config/', api_views.ConfigView.as_view()),
    url(r'^api/api-auth/', include('rest_framework.urls', namespace='rest_framework')),
    url(r'^api/related_fields', staff_member_required(api_views.RelatedFieldsView.as_view()), name="related_fields"),
    url(r'^api/fields', staff_member_required(api_views.FieldsView.as_view()), name="fields"),
    url(r'^api/report/(?P<report_id>\w+)/generate/', staff_member_required(api_views.GenerateReport.as_view()), name="generate_report"),
    url('^report/(?P<pk>\d+)/$', views.ReportSPAView.as_view(), name="report_update_view"),
]

if not hasattr(settings, 'REPORT_BUILDER_FRONTEND') or settings.REPORT_BUILDER_FRONTEND:
    urlpatterns += [
        url(r'^', staff_member_required(views.ReportSPAView.as_view()), name="report_builder"),
    ]
